//-----------------------------------------------------------------------------
// Connection.c
//   Definition of the Python type OracleConnection.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// structure for the Python type "Connection"
//-----------------------------------------------------------------------------
typedef struct {
  PyObject_HEAD
  OCISvcCtx *i_Handle;
  OCIEnv *i_EnvHandle;
  OCIError *i_ErrorHandle;
  PyObject *i_UserName;
  PyObject *i_Password;
  PyObject *i_TNSEntry;
  PyObject *i_Version;
  char i_ErrorBuffer[1024];
  sb4 i_ErrorNumber;
  int i_MaxStringSize;
} udt_Connection;


//-----------------------------------------------------------------------------
// functions for the Python type "Connection"
//-----------------------------------------------------------------------------
static void Connection_Free(udt_Connection*);
static PyObject *Connection_GetAttr(udt_Connection*, PyObject*);
static PyObject *Connection_Close(udt_Connection*, PyObject*);
static PyObject *Connection_Commit(udt_Connection*, PyObject*);
static PyObject *Connection_Rollback(udt_Connection*, PyObject*);
static PyObject *Connection_NewCursor(udt_Connection*, PyObject*);
static PyObject *Connection_Version(udt_Connection*);


//-----------------------------------------------------------------------------
// declaration of Python type "Connection"
//-----------------------------------------------------------------------------
static PyTypeObject g_ConnectionType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "OracleConnection",			// type name string
  sizeof(udt_Connection),		// basic size (memory allocated)
  0,					// item size (variable object)
  (destructor) Connection_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) Connection_GetAttr,	// getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// declaration of methods for Python type "Connection"
//-----------------------------------------------------------------------------
static PyMethodDef g_ConnectionMethods[] = {
  { "close", (PyCFunction) Connection_Close, METH_VARARGS },
  { "commit", (PyCFunction) Connection_Commit, METH_VARARGS },
  { "rollback", (PyCFunction) Connection_Rollback, METH_VARARGS },
  { "cursor", (PyCFunction) Connection_NewCursor, METH_VARARGS },
  { NULL, NULL }
};


//-----------------------------------------------------------------------------
// Connection_RaiseOracleError()
//   Reads the error that was caused by the last Oracle statement and raise an
// exception for Python. At this point it is assumed that the Oracle
// environment is fully initialized.
//-----------------------------------------------------------------------------
static void Connection_RaiseOracleError(
  udt_Connection *a_Connection)		// connection to raise error for
{
  sword v_Status;

  v_Status = OCIErrorGet(a_Connection->i_ErrorHandle, 1, 0,
      &a_Connection->i_ErrorNumber, a_Connection->i_ErrorBuffer,
      sizeof(a_Connection->i_ErrorBuffer), OCI_HTYPE_ERROR);
  if (v_Status == OCI_SUCCESS)
    PyErr_SetString(g_DatabaseErrorException, a_Connection->i_ErrorBuffer);
  else PyErr_SetString(g_InternalErrorException, "No Oracle error?");
}


//-----------------------------------------------------------------------------
// Connection_CheckForError()
//   Check for an error in the last call and if an error has occurred, raise a
// Python exception.
//-----------------------------------------------------------------------------
static int Connection_CheckForError(
  udt_Connection *a_Connection,		// connection to check
  sword a_Status)			// status of last call
{
  if (a_Status != OCI_SUCCESS && a_Status != OCI_SUCCESS_WITH_INFO) {
    if (a_Status == OCI_INVALID_HANDLE)
      PyErr_SetString(g_DatabaseErrorException, "Invalid handle!");
    else Connection_RaiseOracleError(a_Connection);
    return -1;
  }
  return 0;
}


//-----------------------------------------------------------------------------
// Connection_IsConnected()
//   Determines if the connection object is connected to the database. If not,
// a Python exception is raised.
//-----------------------------------------------------------------------------
static int Connection_IsConnected(
  udt_Connection *a_Connection)		// connection to check
{
  if (!a_Connection->i_Handle) {
    PyErr_SetString(g_InterfaceErrorException, "not connected");
    return -1;
  }
  return 0;
}


#include "Cursor.c"


//-----------------------------------------------------------------------------
// Connection_New()
//   Create a new connection object.
//-----------------------------------------------------------------------------
static udt_Connection *Connection_New(void)
{
  udt_Connection *v_Connection;
  sword v_Status;
  ub4 v_Mode;

  // create a new object for the Oracle connection
  v_Connection = PyObject_NEW(udt_Connection, &g_ConnectionType);
  if (!v_Connection)
    return NULL;
  v_Connection->i_Handle = NULL;
  v_Connection->i_EnvHandle = NULL;
  v_Connection->i_ErrorHandle = NULL;
  v_Connection->i_UserName = NULL;
  v_Connection->i_Password = NULL;
  v_Connection->i_TNSEntry = NULL;
  v_Connection->i_Version = NULL;
  v_Connection->i_MaxStringSize = 4000;

  // turn threading mode on, if desired
  v_Mode = OCI_OBJECT;
  if (PyDict_GetItemString(g_ModuleDictionary, "OPT_Threading"))
    v_Mode |= OCI_THREADED | OCI_NO_MUTEX;

  // create the environment handle
  v_Status = OCIEnvCreate(&v_Connection->i_EnvHandle, v_Mode, NULL, NULL, NULL,
      NULL, 0, NULL);
  if (v_Status != OCI_SUCCESS) {
    v_Connection->i_EnvHandle = NULL;
    Py_DECREF(v_Connection);
    PyErr_SetString(PyExc_RuntimeError,
        "Unable to acquire Oracle environment handle");
    return NULL;
  }

  // create the error handle
  v_Status = OCIHandleAlloc(v_Connection->i_EnvHandle,
      (dvoid**) &v_Connection->i_ErrorHandle, OCI_HTYPE_ERROR, 0, 0);
  if (v_Status != OCI_SUCCESS) {
    v_Connection->i_ErrorHandle = NULL;
    Py_DECREF(v_Connection);
    PyErr_SetString(PyExc_RuntimeError,
        "Unable to acquire Oracle error handle");
    return NULL;
  }

  return v_Connection;
}


//-----------------------------------------------------------------------------
// Connection_SetBufferSize()
//   Set the buffer size for the connection.
//-----------------------------------------------------------------------------
static int Connection_SetBufferSize(
  udt_Connection *a_Connection)		// connection to set buffer size on
{
  PyObject *v_Version;

  // ignore checking for Oracle 7 if so desired (done by setting the attribute
  // OPT_NoOracle7 on the module) as this significantly improves performance
  if (PyDict_GetItemString(g_ModuleDictionary, "OPT_NoOracle7"))
    return 0;

  // determine the version of connection
  v_Version = Connection_Version(a_Connection);
  Py_XDECREF(v_Version);
  if (!v_Version) {
    Py_DECREF(a_Connection);
    return -1;
  }

  // Oracle7 connections only handle 2000 bytes
  if (*PyString_AS_STRING(a_Connection->i_Version) == '7')
    a_Connection->i_MaxStringSize = 2000;
  return 0;
}


//-----------------------------------------------------------------------------
// Connection_Connect()
//   Create a new connection object, connecting to the database.
//-----------------------------------------------------------------------------
static PyObject *Connection_Connect(
  const char *a_UserName,		// user name
  unsigned a_UserNameLen,		// length of user name
  const char *a_Password,		// password
  unsigned a_PasswordLen,		// length of password
  const char *a_TNSEntry,		// TNS entry
  unsigned a_TNSEntryLen)		// length of TNS entry
{
  udt_Connection *v_Connection;
  sword v_Status;

  // create a new object for the Oracle connection
  v_Connection = Connection_New();
  if (!v_Connection)
    return NULL;

  // create the string for the username
  v_Connection->i_UserName = PyString_FromStringAndSize(a_UserName,
      a_UserNameLen);
  if (!v_Connection->i_UserName) {
    Py_DECREF(v_Connection);
    return NULL;
  }

  // create the string for the password
  v_Connection->i_Password = PyString_FromStringAndSize(a_Password,
      a_PasswordLen);
  if (!v_Connection->i_Password) {
    Py_DECREF(v_Connection);
    return NULL;
  }

  // create the string for the TNS entry
  v_Connection->i_TNSEntry = PyString_FromStringAndSize(a_TNSEntry,
      a_TNSEntryLen);
  if (!v_Connection->i_TNSEntry) {
    Py_DECREF(v_Connection);
    return NULL;
  }

  // attempt logon to the database
  Py_BEGIN_ALLOW_THREADS
  v_Status = OCILogon(
      v_Connection->i_EnvHandle,	// environment handle
      v_Connection->i_ErrorHandle,	// error handle
      &(v_Connection->i_Handle),	// service handle (OUT)
      (text*) a_UserName,		// username
      a_UserNameLen,			// username string length
      (text*) a_Password,		// password
      a_PasswordLen,			// password string length
      (text*) a_TNSEntry,		// TNS entry
      a_TNSEntryLen);			// TNS entry string length
  Py_END_ALLOW_THREADS
  if (Connection_CheckForError(v_Connection, v_Status) < 0) {
    v_Connection->i_Handle = NULL;
    Py_DECREF(v_Connection);
    return NULL;
  }

  // set the buffer size
  if (Connection_SetBufferSize(v_Connection) < 0)
    return NULL;

  return (PyObject*) v_Connection;
}


//-----------------------------------------------------------------------------
// Connection_Attach()
//   Create a new connection object, attaching to an existing handle.
//-----------------------------------------------------------------------------
static PyObject *Connection_Attach(
  OCISvcCtx *a_Handle)			// handle to attach to
{
  udt_Connection *v_Connection;

  // create a new object for the Oracle connection
  v_Connection = Connection_New();
  if (!v_Connection)
    return NULL;

  // create the strings for the username, password and TNS entry
  Py_INCREF(Py_None);
  v_Connection->i_UserName = Py_None;
  Py_INCREF(Py_None);
  v_Connection->i_Password = Py_None;
  Py_INCREF(Py_None);
  v_Connection->i_TNSEntry = Py_None;

  // set the handle to the one passed in
  v_Connection->i_Handle = a_Handle;

  // set the buffer size
  if (Connection_SetBufferSize(v_Connection) < 0)
    return NULL;

  return (PyObject*) v_Connection;
}


//-----------------------------------------------------------------------------
// Connection_Free()
//   Deallocate the connection, disconnecting from the database if necessary.
//-----------------------------------------------------------------------------
static void Connection_Free(
  udt_Connection *a_Connection)		// connection object
{
  if (a_Connection->i_Handle) {
    Py_BEGIN_ALLOW_THREADS
    OCILogoff(a_Connection->i_Handle, a_Connection->i_ErrorHandle);
    OCIHandleFree(a_Connection->i_Handle, OCI_HTYPE_SVCCTX);
    Py_END_ALLOW_THREADS
  }
  if (a_Connection->i_ErrorHandle)
    OCIHandleFree(a_Connection->i_ErrorHandle, OCI_HTYPE_ERROR);
  if (a_Connection->i_EnvHandle)
    OCIHandleFree(a_Connection->i_EnvHandle, OCI_HTYPE_ENV);
  Py_XDECREF(a_Connection->i_UserName);
  Py_XDECREF(a_Connection->i_Password);
  Py_XDECREF(a_Connection->i_TNSEntry);
  Py_XDECREF(a_Connection->i_Version);
  PyMem_DEL(a_Connection);
}


//-----------------------------------------------------------------------------
// Connection_Version()
//   Retrieve the version of the database and return it. Note that this
// function also places the result in the associated dictionary so it is only
// calculated once.
//-----------------------------------------------------------------------------
static PyObject *Connection_Version(
  udt_Connection *a_Connection)		// connection object
{
  PyObject *v_VersionVar, *v_Temp, *v_Results;
  udt_Cursor *v_Cursor;

  // if version has already been determined, no need to determine again
  if (a_Connection->i_Version) {
    Py_INCREF(a_Connection->i_Version);
    return a_Connection->i_Version;
  }

  // allocate a cursor to retrieve the version
  v_Cursor = (udt_Cursor*) Cursor_New(a_Connection);
  if (!v_Cursor)
    return NULL;

  // set the input sizes for the bind variables
  v_Temp = Py_BuildValue("{s:O,s:O}", "ver", &g_StringVarType, "compat",
      &g_StringVarType);
  if (!v_Temp) {
    Py_DECREF(v_Cursor);
    return NULL;
  }
  v_Results = Cursor_SetInputSizes(v_Cursor, NULL, v_Temp);
  Py_DECREF(v_Temp);
  if (!v_Results) {
    Py_DECREF(v_Cursor);
    return NULL;
  }
  Py_DECREF(v_Results);

  // create the parameters for the function call
  v_Temp = Py_BuildValue("(s)",
      "begin dbms_utility.db_version(:ver, :compat); end;");
  if (!v_Temp) {
    Py_DECREF(v_Cursor);
    return NULL;
  }

  // execute the cursor
  v_Results = Cursor_Execute(v_Cursor, v_Temp, NULL);
  Py_DECREF(v_Temp);
  if (!v_Results) {
    PyErr_Clear();
    Py_DECREF(v_Cursor);
    if (a_Connection->i_ErrorNumber == 6550) {
      a_Connection->i_Version = PyString_FromString("7.x");
      if (a_Connection->i_Version)
        Py_INCREF(a_Connection->i_Version);
    }
    return a_Connection->i_Version;
  }
  Py_DECREF(v_Results);

  // retrieve value
  v_VersionVar = PyDict_GetItemString(v_Cursor->i_BindVariables, "ver");
  if (!v_VersionVar) {
    PyErr_SetString(g_InternalErrorException,
        "unable to locate version variable");
    return NULL;
  }
  a_Connection->i_Version = Variable_GetValue((udt_Variable*) v_VersionVar, 0);
  Py_DECREF(v_Cursor);
  if (!a_Connection->i_Version)
    return NULL;

  Py_INCREF(a_Connection->i_Version);
  return a_Connection->i_Version;
}


//-----------------------------------------------------------------------------
// Connection_GetAttr()
//   Retrieve an attribute on the connection object.
//-----------------------------------------------------------------------------
static PyObject *Connection_GetAttr(
  udt_Connection *a_Connection,		// connection object
  PyObject *a_Name)			// name of attribute
{
  PyObject *v_Attribute;
  char *v_Name;

  v_Name = PyString_AS_STRING(a_Name);
  v_Attribute = Py_FindMethod(g_ConnectionMethods, (PyObject*) a_Connection,
      v_Name);
  if (v_Attribute)
    return v_Attribute;
  PyErr_Clear();
  if (v_Name[0] == 'v' && strcmp(v_Name, "version") == 0)
    return Connection_Version(a_Connection);
  if (v_Name[0] == 'u' && strcmp(v_Name, "username") == 0) {
    Py_INCREF(a_Connection->i_UserName);
    return a_Connection->i_UserName;
  } else if (v_Name[0] == 'p' && strcmp(v_Name, "password") == 0) {
    Py_INCREF(a_Connection->i_Password);
    return a_Connection->i_Password;
  } else if (v_Name[0] == 't' && strcmp(v_Name, "tnsentry") == 0) {
    Py_INCREF(a_Connection->i_TNSEntry);
    return a_Connection->i_TNSEntry;
  }
  PyErr_SetObject(PyExc_AttributeError, a_Name);
  return NULL;
}


//-----------------------------------------------------------------------------
// Connection_Close()
//   Disconnect from the connection.
//-----------------------------------------------------------------------------
static PyObject *Connection_Close(
  udt_Connection *a_Connection,		// connection to close
  PyObject *a_Args)			// arguments
{
  sword v_Status;

  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // make sure we are actually connected
  if (Connection_IsConnected(a_Connection) < 0)
    return NULL;

  // logoff of the server
  Py_BEGIN_ALLOW_THREADS
  v_Status = OCILogoff(a_Connection->i_Handle, a_Connection->i_ErrorHandle);
  Py_END_ALLOW_THREADS
  if (Connection_CheckForError(a_Connection, v_Status) < 0)
    return NULL;
  OCIHandleFree(a_Connection->i_Handle, OCI_HTYPE_SVCCTX);
  a_Connection->i_Handle = NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Connection_Commit()
//   Commit the transaction on the connection.
//-----------------------------------------------------------------------------
static PyObject *Connection_Commit(
  udt_Connection *a_Connection,		// connection to close
  PyObject *a_Args)			// arguments
{
  sword v_Status;

  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // make sure we are actually connected
  if (Connection_IsConnected(a_Connection) < 0)
    return NULL;

  // perform the commit
  Py_BEGIN_ALLOW_THREADS
  v_Status = OCITransCommit(
      a_Connection->i_Handle,		// service handle
      a_Connection->i_ErrorHandle,	// error handle
      OCI_DEFAULT);			// mode
  Py_END_ALLOW_THREADS
  if (Connection_CheckForError(a_Connection, v_Status) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Connection_Rollback()
//   Rollback the transaction on the connection.
//-----------------------------------------------------------------------------
static PyObject *Connection_Rollback(
  udt_Connection *a_Connection,		// connection to close
  PyObject *a_Args)			// arguments
{
  sword v_Status;

  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // make sure we are actually connected
  if (Connection_IsConnected(a_Connection) < 0)
    return NULL;

  // perform the rollback
  Py_BEGIN_ALLOW_THREADS
  v_Status = OCITransRollback(
      a_Connection->i_Handle,		// service handle
      a_Connection->i_ErrorHandle,	// error handle
      OCI_DEFAULT);			// mode
  Py_END_ALLOW_THREADS
  if (Connection_CheckForError(a_Connection, v_Status) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Connection_NewCursor()
//   Create a new cursor (statement) referencing the connection.
//-----------------------------------------------------------------------------
static PyObject *Connection_NewCursor(
  udt_Connection *a_Connection,		// connection to close
  PyObject *a_Args)			// arguments
{
  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  return Cursor_New(a_Connection);
}

