//-----------------------------------------------------------------------------
// Cursor.c
//   Definition of the Python type OracleCursor.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// structure for the Python type "Cursor"
//-----------------------------------------------------------------------------
typedef struct {
  PyObject_HEAD
  OCIStmt *i_Handle;
  udt_Connection *i_Connection;
  PyObject *i_BindVariables;
  PyObject *i_DefineVariables;
  int i_ArraySize;
  int i_SetInputSizes;
  int i_OutputSize;
  int i_OutputSizeColumn;
  int i_RowCount;
  int i_AllocatedRows;
  int i_ActualRows;
  int i_RowNum;
  int i_IsQuery;
} udt_Cursor;


//-----------------------------------------------------------------------------
// macro used to ensure that bus errors do not occur on 64-bit systems
//-----------------------------------------------------------------------------
#define ALLOC_ROWS(x)			((((x) >> 2) + 1) << 2)


//-----------------------------------------------------------------------------
// dependent function defintions
//-----------------------------------------------------------------------------
static void Cursor_Free(udt_Cursor*);


//-----------------------------------------------------------------------------
// functions for the Python type "Cursor"
//-----------------------------------------------------------------------------
static PyObject *Cursor_GetAttr(udt_Cursor*, PyObject*);
static int Cursor_SetAttr(udt_Cursor*, PyObject*, PyObject*);
static PyObject *Cursor_Close(udt_Cursor*, PyObject*);
static PyObject *Cursor_CallProc(udt_Cursor*, PyObject*);
static PyObject *Cursor_Execute(udt_Cursor*, PyObject*, PyObject*);
static PyObject *Cursor_ExecuteMany(udt_Cursor*, PyObject*);
static PyObject *Cursor_ExecuteManyPrepared(udt_Cursor*, PyObject*);
static PyObject *Cursor_FetchOne(udt_Cursor*, PyObject*);
static PyObject *Cursor_FetchMany(udt_Cursor*, PyObject*);
static PyObject *Cursor_FetchAll(udt_Cursor*, PyObject*);
static PyObject *Cursor_FetchRaw(udt_Cursor*, PyObject*);
static PyObject *Cursor_Prepare(udt_Cursor*, PyObject*);
static PyObject *Cursor_SetInputSizes(udt_Cursor*, PyObject*, PyObject*);
static PyObject *Cursor_SetOutputSize(udt_Cursor*, PyObject*);


//-----------------------------------------------------------------------------
// declaration of Python type "Cursor"
//-----------------------------------------------------------------------------
static PyTypeObject g_CursorType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "OracleCursor",			// type name string
  sizeof(udt_Cursor),			// basic size (memory allocated)
  0,					// item size (variable object)
  (destructor) Cursor_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) Cursor_GetAttr,	// getattro function
  (setattrofunc) Cursor_SetAttr,	// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// declaration of methods for Python type "Cursor"
//-----------------------------------------------------------------------------
static PyMethodDef g_CursorMethods[] = {
  { "execute", (PyCFunction) Cursor_Execute, METH_VARARGS | METH_KEYWORDS },
  { "fetchall", (PyCFunction) Cursor_FetchAll, METH_VARARGS },
  { "fetchone", (PyCFunction) Cursor_FetchOne, METH_VARARGS },
  { "fetchmany", (PyCFunction) Cursor_FetchMany, METH_VARARGS },
  { "fetchraw", (PyCFunction) Cursor_FetchRaw, METH_VARARGS },
  { "prepare", (PyCFunction) Cursor_Prepare, METH_VARARGS },
  { "setinputsizes", (PyCFunction) Cursor_SetInputSizes,
        METH_VARARGS | METH_KEYWORDS },
  { "executemany", (PyCFunction) Cursor_ExecuteMany, METH_VARARGS },
  { "callproc", (PyCFunction) Cursor_CallProc, METH_VARARGS },
  { "executemanyprepared", (PyCFunction) Cursor_ExecuteManyPrepared,
        METH_VARARGS },
  { "setoutputsize", (PyCFunction) Cursor_SetOutputSize, METH_VARARGS },
  { "close", (PyCFunction) Cursor_Close, METH_VARARGS },
  { NULL, NULL }
};


#include "Variable.c"


//-----------------------------------------------------------------------------
// Cursor_IsOpen()
//   Determines if the cursor object is open and if so, if the connection is
// also open.
//-----------------------------------------------------------------------------
static int Cursor_IsOpen(
  udt_Cursor *a_Cursor)			// cursor to check
{
  if (!a_Cursor->i_Handle) {
    PyErr_SetString(g_InterfaceErrorException, "not open");
    return -1;
  }
  return Connection_IsConnected(a_Cursor->i_Connection);
}


//-----------------------------------------------------------------------------
// Cursor_New()
//   Create a new cursor object.
//-----------------------------------------------------------------------------
static PyObject *Cursor_New(
  udt_Connection *a_Connection)		// connection to associate with
{
  udt_Cursor *v_Cursor;
  sword v_Status;

  // create a new object for the Oracle connection
  v_Cursor = PyObject_NEW(udt_Cursor, &g_CursorType);
  if (!v_Cursor)
    return NULL;
  v_Cursor->i_Handle = NULL;
  Py_INCREF(a_Connection);
  v_Cursor->i_Connection = a_Connection;
  v_Cursor->i_BindVariables = NULL;
  v_Cursor->i_DefineVariables = NULL;
  v_Cursor->i_ArraySize = 1;
  v_Cursor->i_AllocatedRows = ALLOC_ROWS(v_Cursor->i_ArraySize);
  v_Cursor->i_RowCount = 0;
  v_Cursor->i_SetInputSizes = 0;
  v_Cursor->i_IsQuery = -1;
  v_Cursor->i_OutputSize = -1;
  v_Cursor->i_OutputSizeColumn = -1;

  // allocate an Oracle statement handle
  v_Status = OCIHandleAlloc(a_Connection->i_EnvHandle,
      (dvoid**) &v_Cursor->i_Handle, OCI_HTYPE_STMT, 0, 0);
  if (Connection_CheckForError(a_Connection, v_Status) < 0) {
    Py_DECREF(v_Cursor);
    return NULL;
  }

  return (PyObject*) v_Cursor;
}


//-----------------------------------------------------------------------------
// Cursor_Free()
//   Deallocate the cursor.
//-----------------------------------------------------------------------------
static void Cursor_Free(
  udt_Cursor *a_Cursor)			// cursor object
{
  if (a_Cursor->i_Handle)
    OCIHandleFree(a_Cursor->i_Handle, OCI_HTYPE_STMT);
  Py_XDECREF(a_Cursor->i_BindVariables);
  Py_XDECREF(a_Cursor->i_DefineVariables);
  Py_DECREF(a_Cursor->i_Connection);
  PyMem_DEL(a_Cursor);
}


//-----------------------------------------------------------------------------
// Cursor_PerformDefine()
//   Perform the defines for the cursor. At this point it is assumed that the
// statement being executed is in fact a query.
//-----------------------------------------------------------------------------
static int Cursor_PerformDefine(
  udt_Cursor *a_Cursor)			// cursor to perform define on
{
  int v_NumParams, v_Pos;
  udt_Variable *v_Var;
  sword v_Status;

  // determine number of items in select-list
  v_Status = OCIAttrGet(a_Cursor->i_Handle, OCI_HTYPE_STMT,
      (dvoid*) &v_NumParams, 0, OCI_ATTR_PARAM_COUNT,
      a_Cursor->i_Connection->i_ErrorHandle);
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return -1;

  // create a list corresponding to the number of items
  Py_XDECREF(a_Cursor->i_DefineVariables);
  a_Cursor->i_DefineVariables = PyList_New(v_NumParams);
  if (!a_Cursor->i_DefineVariables)
    return -1;

  // define a variable for each select-item
  for (v_Pos = 1; v_Pos <= v_NumParams; v_Pos++) {
    v_Var = Variable_Define(a_Cursor, v_Pos);
    if (!v_Var)
      return -1;
    PyList_SET_ITEM(a_Cursor->i_DefineVariables, v_Pos - 1,
        (PyObject *) v_Var);
  }

  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_ItemDescription()
//   Return a tuple describing the item at the given position.
//-----------------------------------------------------------------------------
static PyObject *Cursor_ItemDescription(
  udt_Cursor *a_Cursor,			// cursor object
  unsigned a_Pos)			// position
{
  PyObject *v_Tuple, *v_Var, *v_Type;
  int v_DisplaySize, v_Index;
  ub2 v_InternalSize;
  sb2 v_Precision;
  dvoid *v_Param;
  sword v_Status;
  ub4 v_NameLen;
  char *v_Name;
  ub1 v_NullOk;
  sb1 v_Scale;

  // acquire parameter descriptor
  v_Status = OCIParamGet(a_Cursor->i_Handle, OCI_HTYPE_STMT,
      a_Cursor->i_Connection->i_ErrorHandle, &v_Param, a_Pos);
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return NULL;

  // acquire internal size of item
  v_Status = OCIAttrGet(v_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_InternalSize,
      0, OCI_ATTR_DATA_SIZE, a_Cursor->i_Connection->i_ErrorHandle);
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return NULL;

  // aquire name of item
  v_Status = OCIAttrGet(v_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_Name,
      &v_NameLen, OCI_ATTR_NAME, a_Cursor->i_Connection->i_ErrorHandle);
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return NULL;

  // lookup precision and scale
  v_Scale = v_Precision = 0;
  v_Var = PyList_GET_ITEM(a_Cursor->i_DefineVariables, a_Pos - 1);
  if (v_Var->ob_type == &g_NumberVarType) {
    v_Status = OCIAttrGet(v_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_Scale,
        0, OCI_ATTR_SCALE, a_Cursor->i_Connection->i_ErrorHandle);
    if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
      return NULL;
    v_Status = OCIAttrGet(v_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_Precision,
        0, OCI_ATTR_PRECISION, a_Cursor->i_Connection->i_ErrorHandle);
    if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
      return NULL;
  }

  // lookup whether null is permitted for the attribute
  v_Status = OCIAttrGet(v_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_NullOk, 0,
      OCI_ATTR_IS_NULL, a_Cursor->i_Connection->i_ErrorHandle);
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return NULL;

  // set display length based on data type
  v_Type = (PyObject*) v_Var->ob_type;
  if (v_Type == (PyObject*) &g_StringVarType)
    v_DisplaySize = v_InternalSize;
  else if (v_Type == (PyObject*) &g_BinaryVarType)
    v_DisplaySize = v_InternalSize;
  else if (v_Type == (PyObject*) &g_NumberVarType) {
    v_DisplaySize = v_Precision;
    if (v_Scale)
      v_DisplaySize += v_Scale + 1;
  } else if (v_Type == (PyObject*) &g_DateTimeVarType) {
    v_Type = (PyObject*) &g_ExternalDateTimeVarType;
    v_DisplaySize = 23;
  } else if (v_Type == (PyObject*) &g_LobVarType) {
    v_DisplaySize = -1;
    if (((udt_LobVar*) v_Var)->i_DataType == SQLT_CLOB)
      v_Type = (PyObject*) &g_LongStringVarType;
    else v_Type = (PyObject*) &g_LongBinaryVarType;
  } else {
    v_DisplaySize = -1;
  }

  // create the tuple and populate it
  v_Tuple = PyTuple_New(7);
  if (!v_Tuple)
    return NULL;

  // set each of the items in the tuple
  PyTuple_SET_ITEM(v_Tuple, 0, PyString_FromStringAndSize(v_Name, v_NameLen));
  Py_INCREF(v_Type);
  PyTuple_SET_ITEM(v_Tuple, 1, v_Type);
  PyTuple_SET_ITEM(v_Tuple, 2, PyInt_FromLong(v_DisplaySize));
  PyTuple_SET_ITEM(v_Tuple, 3, PyInt_FromLong(v_InternalSize));
  PyTuple_SET_ITEM(v_Tuple, 4, PyInt_FromLong(v_Precision));
  PyTuple_SET_ITEM(v_Tuple, 5, PyInt_FromLong(v_Scale));
  PyTuple_SET_ITEM(v_Tuple, 6, PyInt_FromLong(v_NullOk != 0));

  // make sure the tuple is ok
  for (v_Index = 0; v_Index < 7; v_Index++) {
    if (!PyTuple_GET_ITEM(v_Tuple, v_Index)) {
      Py_DECREF(v_Tuple);
      return NULL;
    }
  }

  return v_Tuple;
}


//-----------------------------------------------------------------------------
// Cursor_Description()
//   Return a list of 7-tuples consisting of the description of the define
// variables.
//-----------------------------------------------------------------------------
static PyObject *Cursor_Description(
  udt_Cursor *a_Cursor)			// cursor object
{
  PyObject *v_Results, *v_Tuple;
  int v_NumItems, v_Index;

  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // if no statement has been executed yet, return None
  if (!a_Cursor->i_DefineVariables) {
    Py_INCREF(Py_None);
    return Py_None;
  }

  // create a list of the required length
  v_NumItems = PyList_GET_SIZE(a_Cursor->i_DefineVariables);
  v_Results = PyList_New(v_NumItems);
  if (!v_Results)
    return NULL;

  // create tuples corresponding to the select-items
  for (v_Index = 0; v_Index < v_NumItems; v_Index++) {
    v_Tuple = Cursor_ItemDescription(a_Cursor, v_Index + 1);
    if (!v_Tuple) {
      Py_DECREF(v_Results);
      return NULL;
    }
    PyList_SET_ITEM(v_Results, v_Index, v_Tuple);
  }

  return v_Results;
}


//-----------------------------------------------------------------------------
// Cursor_GetAttr()
//   Retrieve an attribute on the cursor object.
//-----------------------------------------------------------------------------
static PyObject *Cursor_GetAttr(
  udt_Cursor *a_Cursor,			// cursor object
  PyObject *a_Name)			// name of attribute
{
  PyObject *v_Attribute;
  char *v_Name;

  v_Name = PyString_AS_STRING(a_Name);
  v_Attribute = Py_FindMethod(g_CursorMethods, (PyObject*) a_Cursor, v_Name);
  if (v_Attribute)
    return v_Attribute;
  PyErr_Clear();
  if (v_Name[0] == 'r' && strcmp(v_Name, "rowcount") == 0)
    return PyInt_FromLong(a_Cursor->i_RowCount);
  if (v_Name[0] == 'a' && strcmp(v_Name, "arraysize") == 0)
    return PyInt_FromLong(a_Cursor->i_ArraySize);
  if (v_Name[0] == 'd' && strcmp(v_Name, "description") == 0)
    return Cursor_Description(a_Cursor);
  PyErr_SetObject(PyExc_AttributeError, a_Name);
  return NULL;
}


//-----------------------------------------------------------------------------
// Cursor_SetAttr()
//   Retrieve an attribute on the cursor object.
//-----------------------------------------------------------------------------
static int Cursor_SetAttr(
  udt_Cursor *a_Cursor,			// cursor object
  PyObject *a_Name,			// name of attribute
  PyObject *a_Value)			// value to set
{
  const char *v_Name;

  v_Name = PyString_AS_STRING(a_Name);
  if (v_Name[0] == 'a' && strcmp(v_Name, "arraysize") == 0) {
    if (!PyInt_Check(a_Value)) {
      PyErr_SetString(PyExc_TypeError, "array size must be an integer");
      return -1;
    }
    a_Cursor->i_ArraySize = PyInt_AS_LONG(a_Value);
    a_Cursor->i_AllocatedRows = ALLOC_ROWS(a_Cursor->i_ArraySize);
    return 0;
  }
  PyErr_SetObject(PyExc_AttributeError, a_Name);
  return -1;
}


//-----------------------------------------------------------------------------
// Cursor_Close()
//   Close the cursor.
//-----------------------------------------------------------------------------
static PyObject *Cursor_Close(
  udt_Cursor *a_Cursor,			// cursor to close
  PyObject *a_Args)			// arguments
{
  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // make sure we are actually open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // close the cursor
  OCIHandleFree(a_Cursor->i_Handle, OCI_HTYPE_STMT);
  a_Cursor->i_Handle = NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Cursor_ClearBindVariables()
//   Clear all bind variables for the cursor.
//-----------------------------------------------------------------------------
static int Cursor_ClearBindVariables(
  udt_Cursor *a_Cursor)			// cursor to clear bind variables on
{
  Py_XDECREF(a_Cursor->i_BindVariables);
  a_Cursor->i_BindVariables = PyDict_New();
  if (!a_Cursor->i_BindVariables)
    return -1;
  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_SetBindVariables()
//   Create or set bind variables.
//-----------------------------------------------------------------------------
static int Cursor_SetBindVariables(
  udt_Cursor *a_Cursor,			// cursor to perform binds on
  PyObject *a_Parameters,		// parameters to bind
  unsigned a_ArrayPos)			// array position to set
{
  PyObject *v_Key, *v_Value;
  udt_Variable *v_Var;
  int v_Pos;

  // scan the parameters, creating bind variables, as necessary
  v_Pos = 0;
  while (PyDict_Next(a_Parameters, &v_Pos, &v_Key, &v_Value)) {
    v_Var = (udt_Variable*) PyDict_GetItem(a_Cursor->i_BindVariables, v_Key);
    if (!v_Var) {
      v_Var = Variable_NewByValue(a_Cursor, v_Value,
          a_Cursor->i_AllocatedRows);
      if (!v_Var)
        return -1;
      if (PyDict_SetItem(a_Cursor->i_BindVariables, v_Key,
          (PyObject*) v_Var) < 0) {
        Py_DECREF(v_Var);
        return -1;
      }
      Py_DECREF(v_Var);
    }
    if (Variable_SetValue(v_Var, a_ArrayPos, v_Value) < 0)
      return -1;
  }

  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_PerformBind()
//   Perform the binds on the cursor.
//-----------------------------------------------------------------------------
static int Cursor_PerformBind(
  udt_Cursor *a_Cursor)			// cursor to perform binds on
{
  PyObject *v_Key, *v_Var;
  int v_Pos;

  // set values and perform binds for all bind variables
  if (a_Cursor->i_BindVariables) {
    v_Pos = 0;
    while (PyDict_Next(a_Cursor->i_BindVariables, &v_Pos, &v_Key, &v_Var)) {
      if (Variable_Bind((udt_Variable*) v_Var, a_Cursor, v_Key) < 0)
        return -1;
    }
  }

  // ensure that input sizes are reset
  a_Cursor->i_SetInputSizes = 0;

  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_IsQuery()
//   Determine if the cursor is executing a select statement.
//-----------------------------------------------------------------------------
static int Cursor_IsQuery(
  udt_Cursor *a_Cursor)			// cursor to perform binds on
{
  ub2 v_StatementType;
  sword v_Status;

  v_Status = OCIAttrGet(a_Cursor->i_Handle, OCI_HTYPE_STMT,
      (dvoid*) &v_StatementType, 0, OCI_ATTR_STMT_TYPE,
      a_Cursor->i_Connection->i_ErrorHandle);
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return -1;
  a_Cursor->i_IsQuery = (v_StatementType == OCI_STMT_SELECT);

  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_CreateTuple()
//   Create a tuple consisting of each of the items in the select-list.
//-----------------------------------------------------------------------------
static PyObject *Cursor_CreateTuple(
  udt_Cursor *a_Cursor)			// cursor object
{
  PyObject *v_Tuple, *v_Item;
  int v_NumItems, v_Pos;
  udt_Variable *v_Var;

  // create a new tuple
  v_NumItems = PyList_GET_SIZE(a_Cursor->i_DefineVariables);
  v_Tuple = PyTuple_New(v_NumItems);
  if (!v_Tuple)
    return NULL;

  // acquire the value for each item
  for (v_Pos = 0; v_Pos < v_NumItems; v_Pos++) {
    v_Var = (udt_Variable*) PyList_GET_ITEM(a_Cursor->i_DefineVariables,
        v_Pos);
    v_Item = Variable_GetValue(v_Var, a_Cursor->i_RowNum);
    if (!v_Item) {
      Py_DECREF(v_Tuple);
      return NULL;
    }
    PyTuple_SET_ITEM(v_Tuple, v_Pos, v_Item);
  }

  a_Cursor->i_RowNum++;
  a_Cursor->i_RowCount++;
  return v_Tuple;
}


//-----------------------------------------------------------------------------
// Cursor_InternalPrepare()
//   Internal method for preparing a statement for execution.
//-----------------------------------------------------------------------------
static int Cursor_InternalPrepare(
  udt_Cursor *a_Cursor,			// cursor to perform prepare on
  const char *a_Statement,		// statement to prepare
  unsigned a_StatementLen)		// length of statement to prepare
{
  sword v_Status;

  // prepare statement
  Py_BEGIN_ALLOW_THREADS
  v_Status = OCIStmtPrepare(a_Cursor->i_Handle,
      a_Cursor->i_Connection->i_ErrorHandle, (text*) a_Statement,
      a_StatementLen, OCI_NTV_SYNTAX, OCI_DEFAULT);
  Py_END_ALLOW_THREADS
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return -1;

  // destroy original define variables, if applicable
  if (a_Cursor->i_DefineVariables) {
    Py_DECREF(a_Cursor->i_DefineVariables);
    a_Cursor->i_DefineVariables = NULL;
  }

  // clear bind variables, if applicable
  if (!a_Cursor->i_SetInputSizes && Cursor_ClearBindVariables(a_Cursor) < 0)
    return -1;

  // determine if statement is a query
  if (Cursor_IsQuery(a_Cursor) < 0)
    return -1;

  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_Prepare()
//   Prepare the statement for execution.
//-----------------------------------------------------------------------------
static PyObject *Cursor_Prepare(
  udt_Cursor *a_Cursor,			// cursor to perform prepare on
  PyObject *a_Args)			// arguments
{
  unsigned v_StatementLen;
  const char *v_Statement;

  // statement text is expected
  if (!PyArg_ParseTuple(a_Args, "s#", &v_Statement, &v_StatementLen))
    return NULL;

  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // prepare the statement
  if (Cursor_InternalPrepare(a_Cursor, v_Statement, v_StatementLen) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Cursor_SetRowCount()
//   Set the rowcount variable.
//-----------------------------------------------------------------------------
static int Cursor_SetRowCount(
  udt_Cursor *a_Cursor)			// cursor to set the rowcount on
{
  ub4 v_RowCount;
  sword v_Status;

  if (a_Cursor->i_IsQuery) {
    a_Cursor->i_RowCount = 0;
    a_Cursor->i_ActualRows = -1;
    a_Cursor->i_RowNum = 0;
  } else {
    v_Status = OCIAttrGet(a_Cursor->i_Handle, OCI_HTYPE_STMT, &v_RowCount, 0,
        OCI_ATTR_ROW_COUNT, a_Cursor->i_Connection->i_ErrorHandle);
    if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
      return -1;
    a_Cursor->i_RowCount = v_RowCount;
  }

  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_CallProc()
//   Call a stored procedure and return the (possibly modified) arguments.
//-----------------------------------------------------------------------------
static PyObject *Cursor_CallProc(
  udt_Cursor *a_Cursor,			// cursor to execute
  PyObject *a_Args)			// arguments
{
  PyObject *v_ListOfArguments, *v_Temp, *v_Results, *v_BindVariables;
  int i, v_ProcedureNameLength, v_NumArguments, v_Pos;
  char *v_ProcedureName, *v_Statement, *v_Ptr;
  PyObject *v_Key, *v_Var;
  char v_Buffer[40];

  // expect stored procedure name and optionally a list of arguments
  v_NumArguments = 0;
  v_ListOfArguments = NULL;
  if (!PyArg_ParseTuple(a_Args, "s#|O!", &v_ProcedureName,
      &v_ProcedureNameLength, &PyList_Type, &v_ListOfArguments))
    return NULL;
  if (v_ListOfArguments)
    v_NumArguments = PyList_GET_SIZE(v_ListOfArguments);

  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // create a list of bind variables
  v_BindVariables = PyDict_New();
  if (!v_BindVariables)
    return NULL;
  for (i = 0; i < v_NumArguments; i++) {
    sprintf(v_Buffer, "v%d", i);
    v_Temp = PyString_FromString(v_Buffer);
    if (!v_Temp) {
      Py_DECREF(v_BindVariables);
      return NULL;
    }
    if (PyDict_SetItem(v_BindVariables, v_Temp,
        PyList_GET_ITEM(v_ListOfArguments, i)) < 0) {
      Py_DECREF(v_BindVariables);
      Py_DECREF(v_Temp);
      return NULL;
    }
    Py_DECREF(v_Temp);
  }

  // allocate a string for the statement
  v_Statement = PyMem_Malloc(v_ProcedureNameLength + v_NumArguments * 6 + 15);
  if (!v_Statement) {
    Py_DECREF(v_BindVariables);
    return PyErr_NoMemory();
  }

  // build up the statement
  strcpy(v_Statement, "begin ");
  strcat(v_Statement, v_ProcedureName);
  v_Ptr = v_Statement + strlen(v_Statement);
  *v_Ptr++ = '(';
  for (i = 0; i < v_NumArguments; i++) {
    if (i > 0)
      *v_Ptr++ = ',';
    v_Ptr += sprintf(v_Ptr, ":v%d", i);
  }
  *v_Ptr = '\0';
  strcat(v_Statement, "); end;");

  // create the parameters for the execute call
  v_Temp = Py_BuildValue("(s)", v_Statement);
  PyMem_Free(v_Statement);
  if (!v_Temp) {
    Py_DECREF(v_BindVariables);
    return NULL;
  }

  // execute the cursor
  v_Results = Cursor_Execute(a_Cursor, v_Temp, v_BindVariables);
  Py_DECREF(v_Temp);
  Py_DECREF(v_BindVariables);
  if (!v_Results)
    return NULL;
  Py_DECREF(v_Results);

  // create the return value
  v_Results = PyList_New(v_NumArguments);
  if (!v_Results)
    return NULL;
  v_Pos = 0;
  while (PyDict_Next(a_Cursor->i_BindVariables, &v_Pos, &v_Key, &v_Var)) {
    v_Temp = Variable_GetValue((udt_Variable*) v_Var, 0);
    if (!v_Temp) {
      Py_DECREF(v_Results);
      return NULL;
    }
    i = atoi(PyString_AS_STRING(v_Key) + 1);
    PyList_SET_ITEM(v_Results, i, v_Temp);
  }

  return v_Results;
}


//-----------------------------------------------------------------------------
// Cursor_Execute()
//   Execute the statement.
//-----------------------------------------------------------------------------
static PyObject *Cursor_Execute(
  udt_Cursor *a_Cursor,			// cursor to execute
  PyObject *a_Args,			// arguments
  PyObject *a_Keywords)			// keywords
{
  unsigned v_StatementLen;
  const char *v_Statement;
  sword v_Status;

  // statement text is expected
  if (!PyArg_ParseTuple(a_Args, "z#", &v_Statement, &v_StatementLen))
    return NULL;

  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // prepare the statement, if applicable
  if (v_Statement) {
    if (Cursor_InternalPrepare(a_Cursor, v_Statement, v_StatementLen) < 0)
      return NULL;
  }

  // perform binds
  if (a_Keywords && Cursor_SetBindVariables(a_Cursor, a_Keywords, 0) < 0)
    return NULL;
  if (Cursor_PerformBind(a_Cursor) < 0)
    return NULL;

  // execute the statement
  Py_BEGIN_ALLOW_THREADS
  v_Status = OCIStmtExecute(a_Cursor->i_Connection->i_Handle,
      a_Cursor->i_Handle, a_Cursor->i_Connection->i_ErrorHandle,
      a_Cursor->i_IsQuery ? 0 : 1, 0, 0, 0, OCI_DEFAULT);
  Py_END_ALLOW_THREADS
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return NULL;

  // perform defines, if necessary
  if (a_Cursor->i_IsQuery && !a_Cursor->i_DefineVariables &&
      Cursor_PerformDefine(a_Cursor) < 0)
    return NULL;

  // set the rowcount attribute or the fetch state
  if (Cursor_SetRowCount(a_Cursor) < 0)
    return NULL;

  // for queries, return the defined variables for possible use
  if (a_Cursor->i_IsQuery) {
    Py_INCREF(a_Cursor->i_DefineVariables);
    return a_Cursor->i_DefineVariables;
  }

  // for all other statements, simply return None
  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Cursor_ExecuteMany()
//   Execute the statement many times. The number of times is equivalent to the
// number of elements in the array of dictionaries.
//-----------------------------------------------------------------------------
static PyObject *Cursor_ExecuteMany(
  udt_Cursor *a_Cursor,			// cursor to execute
  PyObject *a_Args)			// arguments
{
  PyObject *v_Arguments, *v_ListOfArguments;
  unsigned v_StatementLen;
  const char *v_Statement;
  sword v_Status;
  int i;

  // expect statement text (optional) plus list of mappings
  v_Arguments = NULL;
  if (!PyArg_ParseTuple(a_Args, "z#O!", &v_Statement, &v_StatementLen,
      &PyList_Type, &v_ListOfArguments))
    return NULL;

  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // prepare the statement, if applicable
  if (v_Statement) {
    if (Cursor_InternalPrepare(a_Cursor, v_Statement, v_StatementLen) < 0)
      return NULL;
  }

  // queries are not supported as the result is undefined
  if (a_Cursor->i_IsQuery) {
    PyErr_SetString(g_NotSupportedErrorException,
        "queries not supported with executemany(): results are undefined");
    return NULL;
  }

  // perform binds
  for (i = 0; i < PyList_GET_SIZE(v_ListOfArguments); i++) {
    v_Arguments = PyList_GET_ITEM(v_ListOfArguments, i);
    if (!PyDict_Check(v_Arguments)) {
      PyErr_SetString(g_InterfaceErrorException,
          "expecting a list of dictionaries");
      return NULL;
    }
    if (Cursor_SetBindVariables(a_Cursor, v_Arguments, i) < 0)
      return NULL;
  }
  if (Cursor_PerformBind(a_Cursor) < 0)
    return NULL;

  // execute the statement
  Py_BEGIN_ALLOW_THREADS
  v_Status = OCIStmtExecute(a_Cursor->i_Connection->i_Handle,
      a_Cursor->i_Handle, a_Cursor->i_Connection->i_ErrorHandle,
      PyList_GET_SIZE(v_ListOfArguments), 0, 0, 0, OCI_DEFAULT);
  Py_END_ALLOW_THREADS
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return NULL;

  // set the rowcount attribute
  if (Cursor_SetRowCount(a_Cursor) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Cursor_ExecuteManyPrepared()
//   Execute the prepared statement the number of times requested. At this
// point, the statement must have been already prepared and the bind variables
// must have their values set.
//-----------------------------------------------------------------------------
static PyObject *Cursor_ExecuteManyPrepared(
  udt_Cursor *a_Cursor,			// cursor to execute
  PyObject *a_Args)			// arguments
{
  int v_NumIters;
  sword v_Status;

  // expect number of times to execute the statement
  if (!PyArg_ParseTuple(a_Args, "i", &v_NumIters))
    return NULL;
  if (v_NumIters > a_Cursor->i_AllocatedRows) {
    PyErr_SetString(g_InterfaceErrorException, "iterations exceed array size");
    return NULL;
  }

  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // queries are not supported as the result is undefined
  if (a_Cursor->i_IsQuery) {
    PyErr_SetString(g_NotSupportedErrorException,
        "queries not supported with executemany(): results are undefined");
    return NULL;
  }

  // perform binds
  if (Cursor_PerformBind(a_Cursor) < 0)
    return NULL;

  // execute the statement
  Py_BEGIN_ALLOW_THREADS
  v_Status = OCIStmtExecute(a_Cursor->i_Connection->i_Handle,
      a_Cursor->i_Handle, a_Cursor->i_Connection->i_ErrorHandle, v_NumIters, 0,
      0, 0, OCI_DEFAULT);
  Py_END_ALLOW_THREADS
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return NULL;

  // set the rowcount attribute
  if (Cursor_SetRowCount(a_Cursor) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Cursor_VerifyFetch()
//   Verify that fetching may happen from this cursor.
//-----------------------------------------------------------------------------
static int Cursor_VerifyFetch(
  udt_Cursor *a_Cursor)			// cursor to fetch from
{
  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return -1;

  // make sure the cursor is for a query
  if (!a_Cursor->i_IsQuery) {
    PyErr_SetString(g_InterfaceErrorException, "not a query");
    return -1;
  }

  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_InternalFetch()
//   Performs the actual fetch from Oracle.
//-----------------------------------------------------------------------------
static int Cursor_InternalFetch(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  int a_NumRowsToFetch)			// number of rows to fetch
{
  sword v_Status;
  ub4 v_RowCount;

  Py_BEGIN_ALLOW_THREADS
  v_Status = OCIStmtFetch(a_Cursor->i_Handle,
      a_Cursor->i_Connection->i_ErrorHandle, a_NumRowsToFetch, OCI_FETCH_NEXT,
      OCI_DEFAULT);
  Py_END_ALLOW_THREADS
  if (v_Status != OCI_NO_DATA &&
      Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return -1;
  v_Status = OCIAttrGet(a_Cursor->i_Handle, OCI_HTYPE_STMT, &v_RowCount, 0,
      OCI_ATTR_ROW_COUNT, a_Cursor->i_Connection->i_ErrorHandle);
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return -1;
  a_Cursor->i_ActualRows = v_RowCount - a_Cursor->i_RowCount;
  a_Cursor->i_RowNum = 0;
  return 0;
}


//-----------------------------------------------------------------------------
// Cursor_MoreRows()
//   Returns an integer indicating if more rows can be retrieved from the
// cursor.
//-----------------------------------------------------------------------------
static int Cursor_MoreRows(
  udt_Cursor *a_Cursor)			// cursor to fetch from
{
  if (a_Cursor->i_RowNum >= a_Cursor->i_ActualRows) {
    if (a_Cursor->i_ActualRows < 0 ||
        a_Cursor->i_ActualRows == a_Cursor->i_AllocatedRows) {
      if (Cursor_InternalFetch(a_Cursor, a_Cursor->i_AllocatedRows) < 0)
        return -1;
    }
    if (a_Cursor->i_RowNum >= a_Cursor->i_ActualRows)
      return 0;
  }
  return 1;
}


//-----------------------------------------------------------------------------
// Cursor_MultiFetch()
//   Return a list consisting of the remaining rows up to the given row limit
// (if specified).
//-----------------------------------------------------------------------------
static PyObject *Cursor_MultiFetch(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  int a_RowLimit)			// row limit
{
  PyObject *v_Results, *v_Tuple;
  int v_Row, v_Status;

  // create an empty list
  v_Results = PyList_New(0);
  if (!v_Results)
    return NULL;

  // fetch as many rows as possible
  for (v_Row = 0; a_RowLimit == 0 || v_Row < a_RowLimit; v_Row++) {
    v_Status = Cursor_MoreRows(a_Cursor);
    if (v_Status < 0) {
      Py_DECREF(v_Results);
      return NULL;
    } else if (v_Status == 0) {
      break;
    } else {
      v_Tuple = Cursor_CreateTuple(a_Cursor);
      if (!v_Tuple) {
        Py_DECREF(v_Results);
        return NULL;
      }
      if (PyList_Append(v_Results, v_Tuple) < 0) {
        Py_DECREF(v_Tuple);
        Py_DECREF(v_Results);
        return NULL;
      }
      Py_DECREF(v_Tuple);
    }
  }

  return v_Results;
}


//-----------------------------------------------------------------------------
// Cursor_FetchOne()
//   Fetch a single row from the cursor.
//-----------------------------------------------------------------------------
static PyObject *Cursor_FetchOne(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  PyObject *a_Args)			// arguments
{
  int v_Status;

  // no arguments are expected
  if (a_Args && !PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // verify fetch can be performed
  if (Cursor_VerifyFetch(a_Cursor) < 0)
    return NULL;

  // setup return value
  v_Status = Cursor_MoreRows(a_Cursor);
  if (v_Status < 0)
    return NULL;
  else if (v_Status > 0)
    return Cursor_CreateTuple(a_Cursor);

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Cursor_FetchMany()
//   Fetch multiple rows from the cursor based on the arraysize.
//-----------------------------------------------------------------------------
static PyObject *Cursor_FetchMany(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  PyObject *a_Args)			// arguments
{
  int v_RowLimit;

  // parse arguments -- optional rowlimit expected
  v_RowLimit = a_Cursor->i_ArraySize;
  if (!PyArg_ParseTuple(a_Args, "|i", &v_RowLimit))
    return NULL;

  // verify fetch can be performed
  if (Cursor_VerifyFetch(a_Cursor) < 0)
    return NULL;

  return Cursor_MultiFetch(a_Cursor, v_RowLimit);
}


//-----------------------------------------------------------------------------
// Cursor_FetchAll()
//   Fetch all remaining rows from the cursor.
//-----------------------------------------------------------------------------
static PyObject *Cursor_FetchAll(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  PyObject *a_Args)			// arguments
{
  // no arguments are expected
  if (!PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // verify fetch can be performed
  if (Cursor_VerifyFetch(a_Cursor) < 0)
    return NULL;

  return Cursor_MultiFetch(a_Cursor, 0);
}


//-----------------------------------------------------------------------------
// Cursor_FetchRaw()
//   Perform raw fetch on the cursor; return the actual number of rows fetched.
//-----------------------------------------------------------------------------
static PyObject *Cursor_FetchRaw(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  PyObject *a_Args)			// arguments
{
  int v_NumRowsToFetch, v_NumRowsFetched;

  // expect an optional number of rows to retrieve
  v_NumRowsToFetch = a_Cursor->i_AllocatedRows;
  if (!PyArg_ParseTuple(a_Args, "|i", &v_NumRowsToFetch))
    return NULL;
  if (v_NumRowsToFetch > a_Cursor->i_AllocatedRows) {
    PyErr_SetString(g_InterfaceErrorException,
        "rows to fetch exceeds array size");
    return NULL;
  }

  // do not attempt to perform fetch if no more rows to fetch
  if (a_Cursor->i_ActualRows > 0 &&
      a_Cursor->i_ActualRows < a_Cursor->i_AllocatedRows)
    return PyInt_FromLong(0);

  // perform internal fetch
  if (Cursor_InternalFetch(a_Cursor, v_NumRowsToFetch) < 0)
    return NULL;

  a_Cursor->i_RowCount += a_Cursor->i_ActualRows;
  v_NumRowsFetched = a_Cursor->i_ActualRows;
  if (a_Cursor->i_ActualRows == v_NumRowsToFetch)
    a_Cursor->i_ActualRows = -1;
  return PyInt_FromLong(v_NumRowsFetched);
}


//-----------------------------------------------------------------------------
// Cursor_SetInputSizes()
//   Set the sizes of the bind variables.
//-----------------------------------------------------------------------------
static PyObject *Cursor_SetInputSizes(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  PyObject *a_Args,			// arguments
  PyObject *a_Keywords)			// keyword arguments
{
  PyObject *v_Key, *v_Value;
  udt_Variable *v_Var;
  int v_Pos;

  // no arguments are expected
  if (a_Args && !PyArg_ParseTuple(a_Args, ""))
    return NULL;

  // make sure the cursor is open
  if (Cursor_IsOpen(a_Cursor) < 0)
    return NULL;

  // eliminate existing bind variables
  if (Cursor_ClearBindVariables(a_Cursor) < 0)
    return NULL;
  a_Cursor->i_SetInputSizes = 1;

  // process each input
  if (a_Keywords) {
    v_Pos = 0;
    while (PyDict_Next(a_Keywords, &v_Pos, &v_Key, &v_Value)) {
      v_Var = Variable_NewByType(a_Cursor, v_Value, a_Cursor->i_AllocatedRows);
      if (!v_Var)
        return NULL;
      if (PyDict_SetItem(a_Cursor->i_BindVariables, v_Key,
          (PyObject*) v_Var) < 0) {
        Py_DECREF(v_Var);
        return NULL;
      }
      Py_DECREF(v_Var);
    }
  }

  Py_INCREF(a_Cursor->i_BindVariables);
  return a_Cursor->i_BindVariables;
}


//-----------------------------------------------------------------------------
// Cursor_SetOutputSize()
//   Set the size of all of the long columns or just one of them.
//-----------------------------------------------------------------------------
static PyObject *Cursor_SetOutputSize(
  udt_Cursor *a_Cursor,			// cursor to fetch from
  PyObject *a_Args)			// arguments
{
  a_Cursor->i_OutputSizeColumn = -1;
  if (!PyArg_ParseTuple(a_Args, "i|i", &a_Cursor->i_OutputSize,
      &a_Cursor->i_OutputSizeColumn))
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}

