//-----------------------------------------------------------------------------
// DateTimeVar.c
//   Defines the routines for handling numeric variables.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// DateTime type
//-----------------------------------------------------------------------------
typedef struct {
  Variable_HEAD
  OCIDate *i_Data;
} udt_DateTimeVar;


#include "ExternalDateTimeVar.c"


//-----------------------------------------------------------------------------
// Declaration of date/time variable functions.
//-----------------------------------------------------------------------------
static int DateTimeVar_Initialize(udt_DateTimeVar*);
static int DateTimeVar_Define(udt_DateTimeVar*, udt_Cursor*, int, OCIParam*);
static int DateTimeVar_Bind(udt_DateTimeVar*, udt_Cursor*, PyObject*);
static int DateTimeVar_SetValue(udt_DateTimeVar*, unsigned, PyObject*);
static PyObject *DateTimeVar_GetValue(udt_DateTimeVar*, unsigned);
static int DateTimeVar_Copy(udt_DateTimeVar*, udt_DateTimeVar*, unsigned,
    unsigned);


//-----------------------------------------------------------------------------
// variable method declaration
//-----------------------------------------------------------------------------
static udt_VariableMethods g_DateTimeVarMethods = {
  (InitializeProc) DateTimeVar_Initialize,
  (FinalizeProc) NULL,
  (DefineProc) DateTimeVar_Define,
  (BindProc) DateTimeVar_Bind,
  (SetValueProc) DateTimeVar_SetValue,
  (GetValueProc) DateTimeVar_GetValue,
  (CopyProc) DateTimeVar_Copy
};


//-----------------------------------------------------------------------------
// Python type declaration
//-----------------------------------------------------------------------------
static PyTypeObject g_DateTimeVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "DateTimeVar",			// type name string
  sizeof(udt_DateTimeVar),		// basic size (memory allocated)
  sizeof(OCIDate) + sizeof(sb2),	// item size (variable object)
  (destructor) Variable_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) Variable_GetAttr,      // getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// DateTimeVar_Initialize()
//   Initialize the variable.
//-----------------------------------------------------------------------------
static int DateTimeVar_Initialize(
  udt_DateTimeVar *a_Var)		// variable to initialize
{
  a_Var->i_Data = (OCIDate*) &(a_Var->i_Indicator[a_Var->i_AllocatedElements]);
  return 0;
}


//-----------------------------------------------------------------------------
// DateTimeVar_Define()
//   Perform the OCI define for the variable.
//-----------------------------------------------------------------------------
static int DateTimeVar_Define(
  udt_DateTimeVar *a_Var,		// variable to initialize
  udt_Cursor *a_Cursor,			// cursor associated with variable
  int a_Position,			// position to define
  OCIParam *a_Param)			// parameter handle
{
  sword v_Status;

  v_Status = OCIDefineByPos(a_Cursor->i_Handle, &a_Var->i_DefineHandle,
      a_Cursor->i_Connection->i_ErrorHandle, a_Position, a_Var->i_Data,
      sizeof(OCIDate),  SQLT_ODT, a_Var->i_Indicator, 0, 0, OCI_DEFAULT);
  return Connection_CheckForError(a_Cursor->i_Connection, v_Status);
}


//-----------------------------------------------------------------------------
// DateTimeVar_Bind()
//   Bind the variable to the given cursor.
//-----------------------------------------------------------------------------
static int DateTimeVar_Bind(
  udt_DateTimeVar *a_Var,		// variable to bind
  udt_Cursor *a_Cursor,			// cursor to perform bind for
  PyObject *a_Name)			// name of bind variable
{
  sword v_Status;

  if (a_Var->i_IsArray) {
    v_Status = OCIBindByName(a_Cursor->i_Handle, &a_Var->i_BindHandle,
        a_Cursor->i_Connection->i_ErrorHandle, PyString_AS_STRING(a_Name),
        PyString_GET_SIZE(a_Name), a_Var->i_Data, sizeof(OCIDate), SQLT_ODT,
        a_Var->i_Indicator, 0, 0, a_Var->i_AllocatedElements,
        &a_Var->i_ActualElements, OCI_DEFAULT);
  } else {
    v_Status = OCIBindByName(a_Cursor->i_Handle, &a_Var->i_BindHandle,
        a_Cursor->i_Connection->i_ErrorHandle, PyString_AS_STRING(a_Name),
        PyString_GET_SIZE(a_Name), a_Var->i_Data, sizeof(OCIDate), SQLT_ODT,
        a_Var->i_Indicator, 0, 0, 0, 0, OCI_DEFAULT);
  }
  return Connection_CheckForError(a_Cursor->i_Connection, v_Status);
}


//-----------------------------------------------------------------------------
// DateTimeVar_SetValue()
//   Set the value of the variable.
//-----------------------------------------------------------------------------
static int DateTimeVar_SetValue(
  udt_DateTimeVar *a_Var,		// variable to set value for
  unsigned a_Pos,			// array position to set
  PyObject *a_Value)			// value to set
{
  udt_ExternalDateTimeVar *v_Value;
  sword v_Status;
  uword v_Valid;

  // make sure a date is being bound
  if (a_Value->ob_type != &g_ExternalDateTimeVarType) {
    PyErr_SetString(PyExc_TypeError, "expecting date data");
    return -1;
  }

  // store a copy of the value
  v_Value = (udt_ExternalDateTimeVar*) a_Value;
  OCIDateSetDate(&a_Var->i_Data[a_Pos], v_Value->i_Year, v_Value->i_Month,
      v_Value->i_Day);
  OCIDateSetTime(&a_Var->i_Data[a_Pos], v_Value->i_Hour, v_Value->i_Minute,
      v_Value->i_Second);
  v_Status = OCIDateCheck(a_Var->i_Connection->i_ErrorHandle,
      &a_Var->i_Data[a_Pos], &v_Valid);
  if (Connection_CheckForError(a_Var->i_Connection, v_Status) < 0)
    return -1;
  if (v_Valid != 0) {
    PyErr_SetString(g_DataErrorException, "invalid date");
    return -1;
  }

  return 0;
}


//-----------------------------------------------------------------------------
// DateTimeVar_GetValue()
//   Returns the value stored at the given array position.
//-----------------------------------------------------------------------------
static PyObject *DateTimeVar_GetValue(
  udt_DateTimeVar *a_Var,		// variable to determine value for
  unsigned a_Pos)			// array position
{
  return ExternalDateTimeVar_NewFromOracle(a_Var, a_Pos);
}


//-----------------------------------------------------------------------------
// DateTimeVar_Copy()
//   Copies the variable value to the target variable.
//-----------------------------------------------------------------------------
static int DateTimeVar_Copy(
  udt_DateTimeVar *a_TargetVar,		// target variable
  udt_DateTimeVar *a_SourceVar,		// source variable
  unsigned a_SourcePos,			// source array position
  unsigned a_TargetPos)			// target array position
{
  memcpy(&a_TargetVar->i_Data[a_TargetPos], &a_SourceVar->i_Data[a_SourcePos],
      sizeof(OCIDate));
  return 0;
}

