//-----------------------------------------------------------------------------
// ExternalDateTimeVar.c
//   Defines the routines for handling date variables external to this module.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// external date/time type
//-----------------------------------------------------------------------------
typedef struct {
  PyObject_HEAD
  unsigned i_Year;
  unsigned i_Month;
  unsigned i_Day;
  unsigned i_Hour;
  unsigned i_Minute;
  unsigned i_Second;
} udt_ExternalDateTimeVar;


//-----------------------------------------------------------------------------
// Declaration of external date variable functions.
//-----------------------------------------------------------------------------
static void ExternalDateTimeVar_Free(udt_ExternalDateTimeVar*);
static PyObject *ExternalDateTimeVar_GetAttr(udt_ExternalDateTimeVar*,
    PyObject*);
static PyObject *ExternalDateTimeVar_Str(udt_ExternalDateTimeVar*);
static int ExternalDateTimeVar_Cmp(udt_ExternalDateTimeVar*,
    udt_ExternalDateTimeVar*);


//-----------------------------------------------------------------------------
// Python type declaration
//-----------------------------------------------------------------------------
static PyTypeObject g_ExternalDateTimeVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "ExternalDateTimeVar",		// type name string
  sizeof(udt_ExternalDateTimeVar),	// basic size (memory allocated)
  0,					// item size (variable object)
  (destructor) ExternalDateTimeVar_Free,
					// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  (cmpfunc) ExternalDateTimeVar_Cmp,	// compare function
  (reprfunc) ExternalDateTimeVar_Str,	// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  (reprfunc) ExternalDateTimeVar_Str,	// as string function
  (getattrofunc) ExternalDateTimeVar_GetAttr,
					// getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// ExternalDateTimeVar_New()
//   Create a new external date variable.
//-----------------------------------------------------------------------------
PyObject *ExternalDateTimeVar_New(
  unsigned a_Year,			// year
  unsigned a_Month,			// month
  unsigned a_Day,			// day
  unsigned a_Hour,			// hour
  unsigned a_Minute,			// minute
  unsigned a_Second)			// second
{
  udt_ExternalDateTimeVar *v_Var;

  v_Var = PyObject_NEW(udt_ExternalDateTimeVar, &g_ExternalDateTimeVarType);
  if (!v_Var)
    return NULL;
  v_Var->i_Year = a_Year;
  v_Var->i_Month = a_Month;
  v_Var->i_Day = a_Day;
  v_Var->i_Hour = a_Hour;
  v_Var->i_Minute = a_Minute;
  v_Var->i_Second = a_Second;

  return (PyObject*) v_Var;
}


//-----------------------------------------------------------------------------
// ExternalDateTimeVar_NewFromOracle()
//   Create a new external date variable.
//-----------------------------------------------------------------------------
PyObject *ExternalDateTimeVar_NewFromOracle(
  udt_DateTimeVar *a_Var,		// variable to fetch data from
  unsigned a_Pos)			// array position to fetch
{
  ub1 v_Hour, v_Minute, v_Second, v_Month, v_Day;
  sb2 v_Year;

  OCIDateGetDate(&a_Var->i_Data[a_Pos], &v_Year, &v_Month, &v_Day);
  OCIDateGetTime(&a_Var->i_Data[a_Pos], &v_Hour, &v_Minute, &v_Second);
  return ExternalDateTimeVar_New(v_Year, v_Month, v_Day, v_Hour, v_Minute,
      v_Second);
}


//-----------------------------------------------------------------------------
// ExternalDateTimeVar_Free()
//   Free an external date variable.
//-----------------------------------------------------------------------------
static void ExternalDateTimeVar_Free(
  udt_ExternalDateTimeVar *a_Var)	// variable to free
{
  PyMem_DEL(a_Var);
}


//-----------------------------------------------------------------------------
// ExternalDateTimeVar_GetAttr()
//   Retrieve an attribute on the external date variable object.
//-----------------------------------------------------------------------------
static PyObject *ExternalDateTimeVar_GetAttr(
  udt_ExternalDateTimeVar *a_Var,	// external date variable object
  PyObject *a_Name)			// name of attribute
{
  const char *v_Name;

  v_Name = PyString_AS_STRING(a_Name);
  if (strcmp(v_Name, "year") == 0)
    return PyInt_FromLong(a_Var->i_Year);
  if (strcmp(v_Name, "month") == 0)
    return PyInt_FromLong(a_Var->i_Month);
  if (strcmp(v_Name, "day") == 0)
    return PyInt_FromLong(a_Var->i_Day);
  if (strcmp(v_Name, "hour") == 0)
    return PyInt_FromLong(a_Var->i_Hour);
  if (strcmp(v_Name, "minute") == 0)
    return PyInt_FromLong(a_Var->i_Minute);
  if (strcmp(v_Name, "second") == 0)
    return PyInt_FromLong(a_Var->i_Second);

  PyErr_SetObject(PyExc_AttributeError, a_Name);
  return NULL;
}


//-----------------------------------------------------------------------------
// ExternalDateTimeVar_Str()
//   Return the string representation of the external date variable object.
//-----------------------------------------------------------------------------
static PyObject *ExternalDateTimeVar_Str(
  udt_ExternalDateTimeVar *a_Var)	// external date variable object
{
  char v_Value[100];

  sprintf(v_Value, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d", a_Var->i_Year,
      a_Var->i_Month, a_Var->i_Day, a_Var->i_Hour, a_Var->i_Minute,
      a_Var->i_Second);
  return PyString_FromString(v_Value);
}

//-----------------------------------------------------------------------------
// ExternalDateTimeVar_Cmp()
//   Return -1 if the second date is less than the first; 0 if the dates are
// the same and +1 if the second date is greater than the first.
//-----------------------------------------------------------------------------
static int ExternalDateTimeVar_Cmp(
  udt_ExternalDateTimeVar *a_Var_1,	// first date
  udt_ExternalDateTimeVar *a_Var_2)	// second date
{
  if (a_Var_1->i_Year < a_Var_2->i_Year)
    return -1;
  if (a_Var_1->i_Year > a_Var_2->i_Year)
    return 1;
  if (a_Var_1->i_Month < a_Var_2->i_Month)
    return -1;
  if (a_Var_1->i_Month > a_Var_2->i_Month)
    return 1;
  if (a_Var_1->i_Day < a_Var_2->i_Day)
    return -1;
  if (a_Var_1->i_Day > a_Var_2->i_Day)
    return 1;
  if (a_Var_1->i_Hour < a_Var_2->i_Hour)
    return -1;
  if (a_Var_1->i_Hour > a_Var_2->i_Hour)
    return 1;
  if (a_Var_1->i_Minute < a_Var_2->i_Minute)
    return -1;
  if (a_Var_1->i_Minute > a_Var_2->i_Minute)
    return 1;
  if (a_Var_1->i_Second < a_Var_2->i_Second)
    return -1;
  if (a_Var_1->i_Second > a_Var_2->i_Second)
    return 1;

  return 0;
}

