//-----------------------------------------------------------------------------
// NumberVar.c
//   Defines the routines for handling numeric variables.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Number type return values
//-----------------------------------------------------------------------------
static int gc_NumberAsUnknown = 0;
static int gc_NumberAsFloat = 1;
static int gc_NumberAsInteger = 2;
static int gc_NumberAsString = 3;

//-----------------------------------------------------------------------------
// Number type
//-----------------------------------------------------------------------------
typedef struct {
  Variable_HEAD
  OCINumber *i_Data;
  int i_ReturnType;
} udt_NumberVar;


//-----------------------------------------------------------------------------
// Declaration of number variable functions.
//-----------------------------------------------------------------------------
static int NumberVar_Initialize(udt_NumberVar*);
static int NumberVar_Define(udt_NumberVar*, udt_Cursor*, int, OCIParam*);
static int NumberVar_Bind(udt_NumberVar*, udt_Cursor*, PyObject*);
static int NumberVar_SetValue(udt_NumberVar*, unsigned, PyObject*);
static PyObject *NumberVar_GetValue(udt_NumberVar*, unsigned);
static int NumberVar_Copy(udt_NumberVar*, udt_NumberVar*, unsigned, unsigned);


//-----------------------------------------------------------------------------
// variable method declaration
//-----------------------------------------------------------------------------
static udt_VariableMethods g_NumberVarMethods = {
  (InitializeProc) NumberVar_Initialize,
  (FinalizeProc) NULL,
  (DefineProc) NumberVar_Define,
  (BindProc) NumberVar_Bind,
  (SetValueProc) NumberVar_SetValue,
  (GetValueProc) NumberVar_GetValue,
  (CopyProc) NumberVar_Copy
};


//-----------------------------------------------------------------------------
// Python type declaration
//-----------------------------------------------------------------------------
static PyTypeObject g_NumberVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "NumberVar",				// type name string
  sizeof(udt_NumberVar),		// basic size (memory allocated)
  sizeof(OCINumber) + sizeof(sb2),	// item size (variable object)
  (destructor) Variable_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) Variable_GetAttr,      // getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// NumberVar_Initialize()
//   Initialize the variable.
//-----------------------------------------------------------------------------
static int NumberVar_Initialize(
  udt_NumberVar *a_Var)			// variable to initialize
{
  a_Var->i_ReturnType = gc_NumberAsUnknown;
  a_Var->i_Data =
      (OCINumber*) &(a_Var->i_Indicator[a_Var->i_AllocatedElements]);
  return 0;
}


//-----------------------------------------------------------------------------
// NumberVar_Define()
//   Perform the OCI define for the variable.
//-----------------------------------------------------------------------------
static int NumberVar_Define(
  udt_NumberVar *a_Var,			// variable to initialize
  udt_Cursor *a_Cursor,			// cursor associated with variable
  int a_Position,			// position to define
  OCIParam *a_Param)			// parameter handle
{
  sb2 v_Precision;
  sword v_Status;
  sb1 v_Scale;

  // determine if number can be treated as an integer
  if (PyDict_GetItemString(g_ModuleDictionary, "OPT_NumbersAsStrings"))
    a_Var->i_ReturnType = gc_NumberAsString;
  else {
    v_Scale = v_Precision = 0;
    v_Status = OCIAttrGet(a_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_Scale, 0,
        OCI_ATTR_SCALE, a_Cursor->i_Connection->i_ErrorHandle);
    if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
      return -1;
    v_Status = OCIAttrGet(a_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_Precision, 0,
        OCI_ATTR_PRECISION, a_Cursor->i_Connection->i_ErrorHandle);
    if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
      return -1;
    if (v_Scale == 0 && v_Precision > 0 && v_Precision < 10)
      a_Var->i_ReturnType = gc_NumberAsInteger;
    else a_Var->i_ReturnType = gc_NumberAsFloat;
  }

  // perform the define
  v_Status = OCIDefineByPos(a_Cursor->i_Handle, &a_Var->i_DefineHandle,
      a_Cursor->i_Connection->i_ErrorHandle, a_Position, a_Var->i_Data,
      sizeof(OCINumber),  SQLT_VNU, a_Var->i_Indicator, 0, 0, OCI_DEFAULT);
  return Connection_CheckForError(a_Cursor->i_Connection, v_Status);
}


//-----------------------------------------------------------------------------
// NumberVar_Bind()
//   Bind the variable to the given cursor.
//-----------------------------------------------------------------------------
static int NumberVar_Bind(
  udt_NumberVar *a_Var,			// variable to bind
  udt_Cursor *a_Cursor,			// cursor to perform bind for
  PyObject *a_Name)			// name of bind variable
{
  sword v_Status;

  if (a_Var->i_IsArray) {
    v_Status = OCIBindByName(a_Cursor->i_Handle, &a_Var->i_BindHandle,
        a_Cursor->i_Connection->i_ErrorHandle, PyString_AS_STRING(a_Name),
        PyString_GET_SIZE(a_Name), a_Var->i_Data, sizeof(OCINumber), SQLT_VNU,
        a_Var->i_Indicator, 0, 0, a_Var->i_AllocatedElements,
        &a_Var->i_ActualElements, OCI_DEFAULT);
  } else {
    v_Status = OCIBindByName(a_Cursor->i_Handle, &a_Var->i_BindHandle,
        a_Cursor->i_Connection->i_ErrorHandle, PyString_AS_STRING(a_Name),
        PyString_GET_SIZE(a_Name), a_Var->i_Data, sizeof(OCINumber), SQLT_VNU,
        a_Var->i_Indicator, 0, 0, 0, 0, OCI_DEFAULT);
  }
  return Connection_CheckForError(a_Cursor->i_Connection, v_Status);
}


//-----------------------------------------------------------------------------
// NumberVar_SetValue()
//   Set the value of the variable.
//-----------------------------------------------------------------------------
static int NumberVar_SetValue(
  udt_NumberVar *a_Var,			// variable to set value for
  unsigned a_Pos,			// array position to set
  PyObject *a_Value)			// value to set
{
  double v_DoubleValue;
  int v_IntegerValue;
  sword v_Status;

  // make sure a number is being bound
  if (!PyInt_Check(a_Value) && !PyFloat_Check(a_Value)) {
    PyErr_SetString(PyExc_TypeError, "expecting numeric data");
    return -1;
  }

  // otherwise, copy the number into the array
  if (PyInt_Check(a_Value)) {
    v_IntegerValue = PyInt_AS_LONG(a_Value);
    v_Status = OCINumberFromInt(a_Var->i_Connection->i_ErrorHandle,
        &v_IntegerValue, sizeof(int), OCI_NUMBER_SIGNED,
        &a_Var->i_Data[a_Pos]);
    if (a_Var->i_ReturnType == gc_NumberAsUnknown)
      a_Var->i_ReturnType = gc_NumberAsInteger;
  } else {
    v_DoubleValue = PyFloat_AS_DOUBLE(a_Value);
    v_Status = OCINumberFromReal(a_Var->i_Connection->i_ErrorHandle,
        &v_DoubleValue, sizeof(double), &a_Var->i_Data[a_Pos]);
    a_Var->i_ReturnType = gc_NumberAsFloat;
  }
  if (Connection_CheckForError(a_Var->i_Connection, v_Status) < 0)
    return -1;

  return 0;
}


//-----------------------------------------------------------------------------
// NumberVar_GetValue()
//   Returns the value stored at the given array position.
//-----------------------------------------------------------------------------
static PyObject *NumberVar_GetValue(
  udt_NumberVar *a_Var,			// variable to determine value for
  unsigned a_Pos)			// array position
{
  char v_StringValue[100];
  double v_DoubleValue;
  int v_IntegerValue;
  ub4 v_StringLength;
  sword v_Status;

  if (a_Var->i_ReturnType == gc_NumberAsInteger) {
    v_Status = OCINumberToInt(a_Var->i_Connection->i_ErrorHandle,
        &a_Var->i_Data[a_Pos], sizeof(int), OCI_NUMBER_SIGNED,
        (dvoid*) &v_IntegerValue);
    if (Connection_CheckForError(a_Var->i_Connection, v_Status) < 0)
      return NULL;
    return PyInt_FromLong(v_IntegerValue);
  }

  if (a_Var->i_ReturnType == gc_NumberAsString) {
    v_StringLength = sizeof(v_StringValue);
    v_Status = OCINumberToText(a_Var->i_Connection->i_ErrorHandle,
        &a_Var->i_Data[a_Pos], "TM9", 3, NULL, 0, &v_StringLength,
        v_StringValue);
    if (Connection_CheckForError(a_Var->i_Connection, v_Status) < 0)
      return NULL;
    return PyString_FromStringAndSize(v_StringValue, v_StringLength);
  }

  v_Status = OCINumberToReal(a_Var->i_Connection->i_ErrorHandle,
      &a_Var->i_Data[a_Pos], sizeof(double), (dvoid*) &v_DoubleValue);
  if (Connection_CheckForError(a_Var->i_Connection, v_Status) < 0)
    return NULL;
  return PyFloat_FromDouble(v_DoubleValue);
}


//-----------------------------------------------------------------------------
// NumberVar_Copy()
//   Copies the variable value to the target variable.
//-----------------------------------------------------------------------------
static int NumberVar_Copy(
  udt_NumberVar *a_TargetVar,		// target variable
  udt_NumberVar *a_SourceVar,		// source variable
  unsigned a_SourcePos,			// source array position
  unsigned a_TargetPos)			// target array position
{
  memcpy(&a_TargetVar->i_Data[a_TargetPos], &a_SourceVar->i_Data[a_SourcePos],
      sizeof(OCINumber));
  return 0;
}

