//-----------------------------------------------------------------------------
// StringVar.c
//   Defines the routines specific to the string type.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Global constants
//-----------------------------------------------------------------------------
const int gc_MaxBufferSize = 32768;
const int gc_DefaultLongStringSize = 128 * 1024;	// 128 KB
const int gc_MaxLongStringSize = 25 * 1024 * 1024;	// 25 MB


//-----------------------------------------------------------------------------
// String type
//-----------------------------------------------------------------------------
typedef struct {
  Variable_HEAD
  int i_MaxLength;
  int i_InIter;
  int i_OutIter;
  char i_Buffer[8192];
  PyObject* i_Data;
  ub2 *i_DummyLength;
  ub2 *i_ReturnCode;
  ub4 *i_ActualLength;
} udt_StringVar;


//-----------------------------------------------------------------------------
// Declaration of string variable functions.
//-----------------------------------------------------------------------------
static int StringVar_Initialize(udt_StringVar*);
static void StringVar_Finalize(udt_StringVar*);
static int StringVar_Define(udt_StringVar*, udt_Cursor*, int, OCIParam*);
static int StringVar_Bind(udt_StringVar*, udt_Cursor*, PyObject*);
static int StringVar_SetValue(udt_StringVar*, unsigned, PyObject*);
static PyObject *StringVar_GetValue(udt_StringVar*, unsigned);
static int StringVar_Copy(udt_StringVar*, udt_StringVar*, unsigned, unsigned);


//-----------------------------------------------------------------------------
// variable method declaration
//-----------------------------------------------------------------------------
static udt_VariableMethods g_StringVarMethods = {
  (InitializeProc) StringVar_Initialize,
  (FinalizeProc) StringVar_Finalize,
  (DefineProc) StringVar_Define,
  (BindProc) StringVar_Bind,
  (SetValueProc) StringVar_SetValue,
  (GetValueProc) StringVar_GetValue,
  (CopyProc) StringVar_Copy
};


//-----------------------------------------------------------------------------
// Python type declarations
//-----------------------------------------------------------------------------
static PyTypeObject g_StringVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "StringVar",				// type name string
  sizeof(udt_StringVar),		// basic size (memory allocated)
  sizeof(sb2) + sizeof(ub2) + sizeof(ub2) + sizeof(ub4),
					// item size (variable object)
  (destructor) Variable_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) Variable_GetAttr,      // getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


static PyTypeObject g_RowidVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "RowidVar",				// type name string
  sizeof(udt_StringVar),		// basic size (memory allocated)
  sizeof(sb2) + sizeof(ub2) + sizeof(ub2) + sizeof(ub4),
					// item size (variable object)
  (destructor) Variable_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) Variable_GetAttr,      // getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


static PyTypeObject g_BinaryVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "BinaryVar",				// type name string
  sizeof(udt_StringVar),		// basic size (memory allocated)
  sizeof(sb2) + sizeof(ub2) + sizeof(ub2) + sizeof(ub4),
					// item size (variable object)
  (destructor) Variable_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) Variable_GetAttr,      // getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


static PyTypeObject g_LongStringVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "LongStringVar",			// type name string
  sizeof(udt_StringVar),		// basic size (memory allocated)
  sizeof(sb2) + sizeof(ub2) + sizeof(ub2) + sizeof(ub4),
					// item size (variable object)
  (destructor) Variable_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) Variable_GetAttr,      // getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


static PyTypeObject g_LongBinaryVarType = {
  PyObject_HEAD_INIT(NULL) 0,		// header initialization
  "LongBinaryVar",			// type name string
  sizeof(udt_StringVar),		// basic size (memory allocated)
  sizeof(sb2) + sizeof(ub2) + sizeof(ub2) + sizeof(ub4),
					// item size (variable object)
  (destructor) Variable_Free,		// deallocation routine
  0,					// print function
  0,					// getattr function
  0,					// setattr function
  0,					// compare function
  0,					// repr function
  0,					// as number function
  0,					// as sequence function
  0,					// as mapping function
  0,					// hash function
  0,					// call function
  0,					// as string function
  (getattrofunc) Variable_GetAttr,      // getattro function
  0,					// setattro function
  0,					// as buffer function
  Py_TPFLAGS_DEFAULT,			// flags
  0					// documentation string
};


//-----------------------------------------------------------------------------
// StringVar_Initialize()
//   Initialize the variable.
//-----------------------------------------------------------------------------
static int StringVar_Initialize(
  udt_StringVar *a_Var)			// variable to initialize
{
  // create list to store string values temporarily
  a_Var->i_Data = PyList_New(a_Var->i_AllocatedElements);
  if (!a_Var->i_Data)
    return -1;

  // initialize members
  a_Var->i_InIter = a_Var->i_OutIter = 0;
  a_Var->i_DummyLength =
      (ub2*) &(a_Var->i_Indicator[a_Var->i_AllocatedElements]);
  a_Var->i_ReturnCode =
      (ub2*) &(a_Var->i_DummyLength[a_Var->i_AllocatedElements]);
  a_Var->i_ActualLength =
      (ub4*) &(a_Var->i_ReturnCode[a_Var->i_AllocatedElements]);

  // set maximum length of variable
  if (a_Var->ob_type == &g_StringVarType || a_Var->ob_type == &g_BinaryVarType)
    a_Var->i_MaxLength = a_Var->i_Connection->i_MaxStringSize;
  else a_Var->i_MaxLength = gc_MaxLongStringSize;

  return 0;
}


//-----------------------------------------------------------------------------
// StringVar_Finalize()
//   Prepare for variable destruction.
//-----------------------------------------------------------------------------
static void StringVar_Finalize(
  udt_StringVar *a_Var)			// variable to free
{
  Py_XDECREF(a_Var->i_Data);
}


//-----------------------------------------------------------------------------
// StringVar_ConcatPiece()
//   Concatenate the current buffer to the current iteration of the fetch.
//-----------------------------------------------------------------------------
int StringVar_ConcatPiece(
  udt_StringVar *a_Var)			// variable to concatenate
{
  PyObject *v_Piece, *v_String;

  if (a_Var->i_ActualLength[a_Var->i_OutIter] == 0)
    return 0;

  v_Piece = PyString_FromStringAndSize(a_Var->i_Buffer,
      a_Var->i_ActualLength[a_Var->i_OutIter]);
  if (!v_Piece)
    return -1;

  v_String = PyList_GET_ITEM(a_Var->i_Data, a_Var->i_OutIter);
  if (v_String) {
    PyString_Concat(&v_String, v_Piece);
    Py_DECREF(v_Piece);
    if (!v_String)
      return -1;
  } else v_String = v_Piece;
  PyList_SET_ITEM(a_Var->i_Data, a_Var->i_OutIter, v_String);

  return 0;
}


//-----------------------------------------------------------------------------
// StringVar_SetBuffer()
//   Set the parameters required for a callback returning data from Oracle to
// the calling program.
//-----------------------------------------------------------------------------
sb4 StringVar_SetBuffer(
  udt_StringVar *a_Var,			// variable to bind
  ub4 a_Iter,				// iteration number
  dvoid **a_Buffer,			// pointer to buffer pointer
  ub4 **a_Length,			// pointer to length pointer
  ub1 *a_Piece,				// piece indicator
  dvoid **a_Indicator,			// pointer to indicator pointer
  ub2 **a_ReturnCode)			// pointer to return code pointer
{
  if (*a_Piece == OCI_FIRST_PIECE || *a_Piece == OCI_ONE_PIECE) {
    if (a_Var->i_OutIter < a_Iter) {
      if (StringVar_ConcatPiece(a_Var) < 0)
        return OCI_ERROR;
    }
    a_Var->i_OutIter = a_Iter;
    Py_XDECREF(PyList_GET_ITEM(a_Var->i_Data, a_Iter));
    PyList_SET_ITEM(a_Var->i_Data, a_Iter, NULL);
  } else if (StringVar_ConcatPiece(a_Var) < 0)
    return OCI_ERROR;

  *a_Buffer = a_Var->i_Buffer;
  a_Var->i_ActualLength[a_Iter] = sizeof(a_Var->i_Buffer);
  *a_Length = &a_Var->i_ActualLength[a_Iter];
#ifdef SQLT_PNTY
  // Oracle 9i seems to work the "right" way
  *a_Indicator = &a_Var->i_Indicator[a_Iter];
  *a_ReturnCode = &a_Var->i_ReturnCode[a_Iter];
#else
  // Oracle 8i has different requirements; does not match documentation
  *a_Indicator = a_Var->i_Indicator;
  *a_ReturnCode = a_Var->i_ReturnCode;
#endif

  if (a_Var->ob_type == &g_StringVarType || a_Var->ob_type == &g_BinaryVarType)
    *a_Piece = OCI_ONE_PIECE;
  else *a_Piece = OCI_NEXT_PIECE;

  return OCI_CONTINUE;
}


//-----------------------------------------------------------------------------
// StringVar_DefineCallback()
//   The callback routine for binding in values from Python strings.
//-----------------------------------------------------------------------------
sb4 StringVar_DefineCallback(
  udt_StringVar *a_Var,			// variable to bind
  OCIDefine *a_DefineHandle,		// define handle
  ub4 a_Iter,				// iteration number
  dvoid **a_Buffer,			// pointer to buffer pointer
  ub4 **a_Length,			// pointer to length pointer
  ub1 *a_Piece,				// piece indicator
  dvoid **a_Indicator,			// pointer to indicator pointer
  ub2 **a_ReturnCode)			// pointer to return code pointer
{
  return StringVar_SetBuffer(a_Var, a_Iter, a_Buffer, a_Length, a_Piece,
      a_Indicator, a_ReturnCode);
}


//-----------------------------------------------------------------------------
// StringVar_Define()
//   Perform the OCI define for the variable.
//-----------------------------------------------------------------------------
static int StringVar_Define(
  udt_StringVar *a_Var,			// variable to initialize
  udt_Cursor *a_Cursor,			// cursor associated with variable
  int a_Position,			// position to define
  OCIParam *a_Param)			// parameter handle
{
  sword v_Status;

  // set the output size, if applicable
  if (a_Cursor->i_OutputSize >= 0) {
    if (a_Cursor->i_OutputSizeColumn < 0 ||
        a_Position == a_Cursor->i_OutputSizeColumn)
      a_Var->i_MaxLength = a_Cursor->i_OutputSize;
  }

  // perform the define
  v_Status = OCIDefineByPos(a_Cursor->i_Handle, &a_Var->i_DefineHandle,
      a_Cursor->i_Connection->i_ErrorHandle, a_Position, 0, a_Var->i_MaxLength,
      a_Var->i_DataType, a_Var->i_Indicator, a_Var->i_DummyLength,
      a_Var->i_ReturnCode, OCI_DYNAMIC_FETCH);
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return -1;
  v_Status = OCIDefineDynamic(a_Var->i_DefineHandle,
      a_Cursor->i_Connection->i_ErrorHandle, a_Var,
      (OCICallbackDefine) StringVar_DefineCallback);
  return Connection_CheckForError(a_Cursor->i_Connection, v_Status);
}


//-----------------------------------------------------------------------------
// StringVar_BindInCallback()
//   The callback routine for binding in values from Python strings.
//-----------------------------------------------------------------------------
sb4 StringVar_BindInCallback(
  udt_StringVar *a_Var,			// variable to bind
  OCIBind *a_BindHandle,		// bind handle
  ub4 a_Iter,				// iteration number
  ub4 a_Index,				// index into array
  dvoid **a_Buffer,			// pointer to buffer pointer
  ub4 *a_Length,			// pointer to length
  ub1 *a_Piece,				// piece indicator
  dvoid **a_Indicator)			// pointer to indicator pointer
{
  PyObject *v_Value;
  char *v_Buffer;
  int v_Length;

  if (a_Var->i_IsDirectlyBound) {
    v_Value = StringVar_GetValue(a_Var, a_Index + a_Iter);
    Py_DECREF(v_Value);
  } else {
    v_Value = PyList_GET_ITEM(a_Var->i_Data, a_Index + a_Iter);
  }
  v_Length = PyString_GET_SIZE(v_Value);
  v_Buffer = PyString_AS_STRING(v_Value);
  if (a_Var->i_InIter == 0) {
    *a_Buffer = v_Buffer;
    if (v_Length > gc_MaxBufferSize) {
      *a_Length = gc_MaxBufferSize;
      *a_Piece = OCI_FIRST_PIECE;
      a_Var->i_InIter++;
    } else {
      *a_Length = v_Length;
      *a_Piece = OCI_ONE_PIECE;
    }
  } else {
    *a_Buffer = v_Buffer + a_Var->i_InIter * gc_MaxBufferSize;
    v_Length -= a_Var->i_InIter * gc_MaxBufferSize;
    if (v_Length > gc_MaxBufferSize) {
      *a_Length = gc_MaxBufferSize;
      *a_Piece = OCI_NEXT_PIECE;
      a_Var->i_InIter++;
    } else {
      *a_Length = v_Length;
      *a_Piece = OCI_LAST_PIECE;
      a_Var->i_InIter = 0;
    }
  }
  *a_Indicator = NULL;

  return OCI_CONTINUE;
}


//-----------------------------------------------------------------------------
// StringVar_BindOutCallback()
//   The callback routine for binding out values to Python strings.
//-----------------------------------------------------------------------------
sb4 StringVar_BindOutCallback(
  udt_StringVar *a_Var,			// variable to bind
  OCIBind *a_BindHandle,		// bind handle
  ub4 a_Iter,				// iteration number
  ub4 a_Index,				// index into array
  dvoid **a_Buffer,			// pointer to buffer pointer
  ub4 **a_Length,			// pointer to length
  ub1 *a_Piece,				// piece indicator
  dvoid **a_Indicator,			// pointer to indicator pointer
  ub2 **a_ReturnCode)			// pointer to return code pointer
{
  return StringVar_SetBuffer(a_Var, a_Index + a_Iter, a_Buffer, a_Length,
      a_Piece, a_Indicator, a_ReturnCode);
}


//-----------------------------------------------------------------------------
// StringVar_Bind()
//   Bind the variable to the given cursor.
//-----------------------------------------------------------------------------
static int StringVar_Bind(
  udt_StringVar *a_Var,			// variable to bind
  udt_Cursor *a_Cursor,			// cursor to perform bind for
  PyObject *a_Name)			// name of bind variable
{
  sword v_Status;

  if (a_Var->i_IsArray) {
    v_Status = OCIBindByName(a_Cursor->i_Handle, &a_Var->i_BindHandle,
        a_Cursor->i_Connection->i_ErrorHandle, PyString_AS_STRING(a_Name),
        PyString_GET_SIZE(a_Name), 0, a_Var->i_MaxLength, a_Var->i_DataType,
        a_Var->i_Indicator, a_Var->i_DummyLength, a_Var->i_ReturnCode,
        a_Var->i_AllocatedElements, &a_Var->i_ActualElements,
        OCI_DATA_AT_EXEC);
  } else {
    v_Status = OCIBindByName(a_Cursor->i_Handle, &a_Var->i_BindHandle,
        a_Cursor->i_Connection->i_ErrorHandle, PyString_AS_STRING(a_Name),
        PyString_GET_SIZE(a_Name), 0, a_Var->i_MaxLength, a_Var->i_DataType,
        a_Var->i_Indicator, a_Var->i_DummyLength, a_Var->i_ReturnCode, 0, 0,
        OCI_DATA_AT_EXEC);
  }
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return -1;
  v_Status = OCIBindDynamic(a_Var->i_BindHandle,
      a_Cursor->i_Connection->i_ErrorHandle, a_Var,
      (OCICallbackInBind) StringVar_BindInCallback, a_Var,
      (OCICallbackOutBind) StringVar_BindOutCallback);
  return Connection_CheckForError(a_Cursor->i_Connection, v_Status);
}


//-----------------------------------------------------------------------------
// StringVar_SetValue()
//   Set the value of the variable.
//-----------------------------------------------------------------------------
static int StringVar_SetValue(
  udt_StringVar *a_Var,			// variable to set value for
  unsigned a_Pos,			// array position to set
  PyObject *a_Value)			// value to set
{
  // make sure a string is being bound
  if (!PyString_Check(a_Value)) {
    PyErr_SetString(PyExc_TypeError, "expecting string data");
    return -1;

  // zero length is null as far as Oracle is concerned
  } else if (PyString_GET_SIZE(a_Value) == 0) {
    a_Var->i_Indicator[a_Pos] = OCI_IND_NULL;

  // otherwise, keep a copy of the string
  } else {
    a_Var->i_ReturnCode[a_Pos] = 0;
    a_Var->i_ActualLength[a_Pos] = 0;
    a_Var->i_DummyLength[a_Pos] = 65535;
    Py_XDECREF(PyList_GET_ITEM(a_Var->i_Data, a_Pos));
    Py_INCREF(a_Value);
    PyList_SET_ITEM(a_Var->i_Data, a_Pos, a_Value);
  }

  return 0;
}


//-----------------------------------------------------------------------------
// StringVar_GetValue()
//   Returns the value stored at the given array position.
//-----------------------------------------------------------------------------
static PyObject *StringVar_GetValue(
  udt_StringVar *a_Var,			// variable to determine value for
  unsigned a_Pos)			// array position
{
  PyObject *v_String;

  if (a_Var->i_ReturnCode[a_Pos] != 0) {
    char v_Buffer[100];
    sprintf(v_Buffer, "column at array pos %d fetched with error: %d",
        a_Pos, a_Var->i_ReturnCode[a_Pos]);
    PyErr_SetString(g_DatabaseErrorException, v_Buffer);
    return NULL;
  }

  if (a_Pos == a_Var->i_OutIter) {
    if (StringVar_ConcatPiece(a_Var) < 0)
      return NULL;
    a_Var->i_OutIter++;
  }

  v_String = PyList_GET_ITEM(a_Var->i_Data, a_Pos);
  Py_INCREF(v_String);
  return v_String;
}


//-----------------------------------------------------------------------------
// StringVar_Copy()
//   Copies the variable value to the target variable.
//-----------------------------------------------------------------------------
static int StringVar_Copy(
  udt_StringVar *a_TargetVar,		// target variable
  udt_StringVar *a_SourceVar,		// source variable
  unsigned a_SourcePos,			// source array position
  unsigned a_TargetPos)			// target array position
{
  PyObject *v_String;

  v_String = StringVar_GetValue(a_SourceVar, a_SourcePos);
  if (!v_String)
    return -1;
  Py_DECREF(v_String);
  return StringVar_SetValue(a_TargetVar, a_TargetPos, v_String);
}

