//-----------------------------------------------------------------------------
// Variable.c
//   Defines Python types for Oracle variables.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// define structure common to all variables
//-----------------------------------------------------------------------------
struct _udt_VariableMethods;
#define Variable_HEAD \
    PyObject_VAR_HEAD \
    OCIBind *i_BindHandle; \
    OCIDefine *i_DefineHandle; \
    udt_Connection *i_Connection; \
    ub4 i_AllocatedElements; \
    ub4 i_ActualElements; \
    int i_IsArray; \
    int i_IsDirectlyBound; \
    sb2 *i_Indicator; \
    ub2 i_DataType; \
    struct _udt_VariableMethods *i_Methods;
typedef struct {
  Variable_HEAD
} udt_Variable;


//-----------------------------------------------------------------------------
// define function types for the common actions that take place on a variable
//-----------------------------------------------------------------------------
typedef int (*InitializeProc)(udt_Variable*, udt_Cursor*);
typedef void (*FinalizeProc)(udt_Variable*);
typedef int (*DefineProc)(udt_Variable*, udt_Cursor*, int, OCIParam*);
typedef int (*BindProc)(udt_Variable*, udt_Cursor*, PyObject*);
typedef int (*SetValueProc)(udt_Variable*, unsigned, PyObject*);
typedef PyObject * (*GetValueProc)(udt_Variable*, unsigned);
typedef int (*CopyProc)(udt_Variable*, udt_Variable*, unsigned, unsigned);


//-----------------------------------------------------------------------------
// define structure for the common actions that take place on a variable
//-----------------------------------------------------------------------------
typedef struct _udt_VariableMethods {
  InitializeProc i_InitializeProc;
  FinalizeProc i_FinalizeProc;
  DefineProc i_DefineProc;
  BindProc i_BindProc;
  SetValueProc i_SetValueProc;
  GetValueProc i_GetValueProc;
  CopyProc i_CopyProc;
} udt_VariableMethods;


//-----------------------------------------------------------------------------
// Declaration of common variable functions.
//-----------------------------------------------------------------------------
static void Variable_Free(udt_Variable *);
static PyObject *Variable_GetAttr(udt_Variable *, PyObject *);
static PyObject *Variable_Copy(udt_Variable *, PyObject *);
static PyObject *Variable_ExternalSetValue(udt_Variable *, PyObject *);
static PyObject *Variable_ExternalGetValue(udt_Variable *, PyObject *);


//-----------------------------------------------------------------------------
// declaration of methods for variables
//-----------------------------------------------------------------------------
static PyMethodDef g_VariableMethods[] = {
  { "copy", (PyCFunction) Variable_Copy, METH_VARARGS },
  { "setvalue", (PyCFunction) Variable_ExternalSetValue, METH_VARARGS },
  { "getvalue", (PyCFunction) Variable_ExternalGetValue, METH_VARARGS },
  { NULL, NULL }
};


#include "StringVar.c"
#include "NumberVar.c"
#include "DateTimeVar.c"
#include "LobVar.c"
#include "CursorVar.c"


//-----------------------------------------------------------------------------
// Variable_New()
//   Allocate a new variable.
//-----------------------------------------------------------------------------
static udt_Variable *Variable_New(
  udt_Cursor *a_Cursor,			// cursor to associate variable with
  unsigned a_NumElements,		// number of elements to allocate
  PyTypeObject *a_TypeObject,		// type of object to create
  ub2 a_OracleDataType,			// Oracle data type to associate
  udt_VariableMethods *a_Methods)	// methods for the variable
{
  udt_Variable *v_Var;
  int i;

  // attempt to allocate the object
  v_Var = PyObject_NEW_VAR(udt_Variable, a_TypeObject, a_NumElements);
  if (!v_Var)
    return NULL;

  // perform basic initialization
  Py_INCREF(a_Cursor->i_Connection);
  v_Var->i_Connection = a_Cursor->i_Connection;
  v_Var->i_BindHandle = NULL;
  v_Var->i_DefineHandle = NULL;
  v_Var->i_AllocatedElements = a_NumElements;
  v_Var->i_ActualElements = 0;
  v_Var->i_IsArray = 0;
  v_Var->i_IsDirectlyBound = 0;
  v_Var->i_DataType = a_OracleDataType;
  v_Var->i_Methods = a_Methods;
  v_Var->i_Indicator = (sb2*) (((char*) v_Var) + a_TypeObject->tp_basicsize);
  for (i = 0; i < a_NumElements; i++)
    v_Var->i_Indicator[i] = OCI_IND_NULL;

  // perform extended initialization
  if ((*v_Var->i_Methods->i_InitializeProc)(v_Var, a_Cursor) < 0) {
    Py_DECREF(v_Var);
    return NULL;
  }

  return v_Var;
}


//-----------------------------------------------------------------------------
// Variable_Free()
//   Free an existing variable.
//-----------------------------------------------------------------------------
static void Variable_Free(
  udt_Variable *a_Var)			// variable to free
{
  if (a_Var->i_Methods->i_FinalizeProc)
    (*a_Var->i_Methods->i_FinalizeProc)(a_Var);
  Py_DECREF(a_Var->i_Connection);
  PyMem_DEL(a_Var);
}


//-----------------------------------------------------------------------------
// Variable_NewByOracleDataType()
//   Allocate a new variable by looking at the Oracle data type.
//-----------------------------------------------------------------------------
static udt_Variable *Variable_NewByOracleDataType(
  udt_Cursor *a_Cursor,			// cursor to associate variable with
  ub2 a_OracleDataType)			// Oracle data type to associate
{
  char v_Buffer[100];

  switch(a_OracleDataType) {
    case SQLT_LNG:
      return Variable_New(a_Cursor, a_Cursor->i_AllocatedRows,
          &g_LongStringVarType, a_OracleDataType, &g_StringVarMethods);
    case SQLT_AFC:
    case SQLT_CHR:
      return Variable_New(a_Cursor, a_Cursor->i_AllocatedRows,
          &g_StringVarType, SQLT_CHR, &g_StringVarMethods);
    case SQLT_RDD:
      return Variable_New(a_Cursor, a_Cursor->i_AllocatedRows,
          &g_RowidVarType, SQLT_CHR, &g_StringVarMethods);
    case SQLT_BIN:
      return Variable_New(a_Cursor, a_Cursor->i_AllocatedRows,
          &g_BinaryVarType, a_OracleDataType, &g_StringVarMethods);
    case SQLT_LBI:
      return Variable_New(a_Cursor, a_Cursor->i_AllocatedRows,
          &g_LongBinaryVarType, a_OracleDataType, &g_StringVarMethods);
    case SQLT_NUM:
      return Variable_New(a_Cursor, a_Cursor->i_AllocatedRows,
          &g_NumberVarType, a_OracleDataType, &g_NumberVarMethods);
    case SQLT_DAT:
      return Variable_New(a_Cursor, a_Cursor->i_AllocatedRows,
          &g_DateTimeVarType, a_OracleDataType, &g_DateTimeVarMethods);
    case SQLT_CLOB:
    case SQLT_BLOB:
      return Variable_New(a_Cursor, a_Cursor->i_AllocatedRows,
          &g_LobVarType, a_OracleDataType, &g_LobVarMethods);
  }

  sprintf(v_Buffer, "Variable_NewByOracleDataType: unhandled data type %d",
      a_OracleDataType);
  PyErr_SetString(g_NotSupportedErrorException, v_Buffer);
  return NULL;
}


//-----------------------------------------------------------------------------
// Variable_NewByValue()
//   Allocate a new variable by looking at the type of the data.
//-----------------------------------------------------------------------------
static udt_Variable *Variable_NewByValue(
  udt_Cursor *a_Cursor,			// cursor to associate variable with
  PyObject *a_Value,			// Python value to associate
  unsigned a_NumElements)		// number of elements to allocate
{
  udt_Variable *v_Var;
  char v_Buffer[200];
  PyObject *v_Value;
  int i;

  // handle scalars
  if (a_Value == Py_None || PyString_Check(a_Value))
    return Variable_New(a_Cursor, a_NumElements, &g_StringVarType, SQLT_CHR,
        &g_StringVarMethods);
  if (PyInt_Check(a_Value) || PyFloat_Check(a_Value)) {
    return Variable_New(a_Cursor, a_NumElements, &g_NumberVarType, SQLT_NUM,
        &g_NumberVarMethods);
  }
  if (a_Value->ob_type == &g_ExternalDateTimeVarType)
    return Variable_New(a_Cursor, a_NumElements, &g_DateTimeVarType, SQLT_DAT,
        &g_DateTimeVarMethods);
  if (a_Value->ob_type == &g_CursorType)
    return Variable_New(a_Cursor, a_NumElements, &g_CursorVarType, SQLT_RSET,
        &g_CursorVarMethods);

  // handle arrays
  if (PyList_Check(a_Value)) {
    v_Value = Py_None;
    for (i = 0; i < PyList_GET_SIZE(a_Value); i++) {
      v_Value = PyList_GET_ITEM(a_Value, i);
      if (v_Value != Py_None)
        break;
    }
    v_Var = Variable_NewByValue(a_Cursor, v_Value,
        ALLOC_ROWS(PyList_GET_SIZE(a_Value)));
    if (!v_Var)
      return NULL;
    v_Var->i_IsArray = 1;
    return v_Var;
  }

  sprintf(v_Buffer, "Variable_NewByValue: unhandled data type %*s",
      150, a_Value->ob_type->tp_name);
  PyErr_SetString(g_NotSupportedErrorException, v_Buffer);
  return NULL;
}


//-----------------------------------------------------------------------------
// Variable_NewArrayByType()
//   Allocate a new PL/SQL array by looking at the Python data type.
//-----------------------------------------------------------------------------
static udt_Variable *Variable_NewArrayByType(
  udt_Cursor *a_Cursor,			// cursor to bind variable to
  PyObject *a_Value)			// value to bind
{
  PyObject *v_Type, *v_Length;
  unsigned v_NumElements;
  udt_Variable *v_Var;

  if (PyList_GET_SIZE(a_Value) != 2) {
    PyErr_SetString(g_ProgrammingErrorException,
        "expecting an array of two elements [type, numelems]");
    return NULL;
  }

  v_Type = PyList_GET_ITEM(a_Value, 0);
  v_Length = PyList_GET_ITEM(a_Value, 1);
  if (!PyInt_Check(v_Length)) {
    PyErr_SetString(g_ProgrammingErrorException,
        "number of elements must be an integer");
    return NULL;
  }

  v_NumElements = ALLOC_ROWS(PyInt_AS_LONG(v_Length));
  if (v_Type == (PyObject*) &g_StringVarType)
    v_Var = Variable_New(a_Cursor, v_NumElements, &g_StringVarType, SQLT_CHR,
        &g_StringVarMethods);
  else if (v_Type == (PyObject*) &g_BinaryVarType)
    v_Var = Variable_New(a_Cursor, v_NumElements, &g_BinaryVarType, SQLT_BIN,
        &g_StringVarMethods);
  else if (v_Type == (PyObject*) &g_NumberVarType)
    v_Var = Variable_New(a_Cursor, v_NumElements, &g_NumberVarType, SQLT_NUM,
        &g_NumberVarMethods);
  else if (v_Type == (PyObject*) &g_ExternalDateTimeVarType)
    v_Var = Variable_New(a_Cursor, v_NumElements, &g_DateTimeVarType, SQLT_DAT,
        &g_DateTimeVarMethods);
  else {
    PyErr_SetString(g_ProgrammingErrorException,
        "type must be a data type supported by cx_Oracle");
    return NULL;
  }
  if (!v_Var)
    return NULL;
  v_Var->i_IsArray = 1;
  return v_Var;
}


//-----------------------------------------------------------------------------
// Variable_NewByType()
//   Allocate a new variable by looking at the Python data type.
//-----------------------------------------------------------------------------
static udt_Variable *Variable_NewByType(
  udt_Cursor *a_Cursor,			// cursor to associate variable with
  PyObject *a_Value,			// Python data type to associate
  unsigned a_NumElements)		// number of elements to allocate
{
  // handle scalars
  if (PyInt_Check(a_Value))
    return Variable_New(a_Cursor, a_NumElements, &g_StringVarType, SQLT_CHR,
        &g_StringVarMethods);
  if (a_Value == (PyObject*) &g_StringVarType ||
      a_Value == (PyObject*) &g_LongStringVarType ||
      a_Value == (PyObject*) &g_RowidVarType)
    return Variable_New(a_Cursor, a_NumElements, (PyTypeObject *) a_Value,
        SQLT_CHR, &g_StringVarMethods);
  if (a_Value == (PyObject*) &g_BinaryVarType)
    return Variable_New(a_Cursor, a_NumElements, (PyTypeObject *) a_Value,
        SQLT_BIN, &g_StringVarMethods);
  if (a_Value == (PyObject*) &g_LongBinaryVarType)
    return Variable_New(a_Cursor, a_NumElements, (PyTypeObject *) a_Value,
        SQLT_LBI, &g_StringVarMethods);
  if (a_Value == (PyObject*) &g_NumberVarType)
    return Variable_New(a_Cursor, a_NumElements, &g_NumberVarType, SQLT_NUM,
        &g_NumberVarMethods);
  if (a_Value == (PyObject*) &g_ExternalDateTimeVarType)
    return Variable_New(a_Cursor, a_NumElements, &g_DateTimeVarType, SQLT_DAT,
        &g_DateTimeVarMethods);

  // handle arrays
  if (PyList_Check(a_Value))
    return Variable_NewArrayByType(a_Cursor, a_Value);

  // handle directly bound variables
  if (a_Value->ob_type == &g_StringVarType ||
      a_Value->ob_type == &g_RowidVarType ||
      a_Value->ob_type == &g_BinaryVarType ||
      a_Value->ob_type == &g_NumberVarType ||
      a_Value->ob_type == &g_DateTimeVarType) {
    Py_INCREF(a_Value);
    ((udt_Variable*) a_Value)->i_IsDirectlyBound = 1;
    return (udt_Variable*) a_Value;
  }

  // invalid input -- raise exception
  PyErr_SetString(g_NotSupportedErrorException,
      "unhandled data type (bind no value)");
  return NULL;
}


//-----------------------------------------------------------------------------
// Variable_Define()
//   Allocate a variable and define it for the given statement.
//-----------------------------------------------------------------------------
static udt_Variable *Variable_Define(
  udt_Cursor *a_Cursor,			// cursor to define for
  int a_Position)			// position to define
{
  udt_Variable *v_Var;
  dvoid *v_Param;
  ub2 v_DataType;
  sword v_Status;

  // retrieve parameter descriptor
  v_Status = OCIParamGet(a_Cursor->i_Handle, OCI_HTYPE_STMT,
      a_Cursor->i_Connection->i_ErrorHandle, &v_Param, a_Position);
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return NULL;

  // retrieve datatype of the parameter
  v_Status = OCIAttrGet(v_Param, OCI_HTYPE_DESCRIBE, (dvoid*) &v_DataType,
      0, OCI_ATTR_DATA_TYPE, a_Cursor->i_Connection->i_ErrorHandle);
  if (Connection_CheckForError(a_Cursor->i_Connection, v_Status) < 0)
    return NULL;

  // create a variable of the correct type
  v_Var = Variable_NewByOracleDataType(a_Cursor, v_DataType);
  if (!v_Var)
    return NULL;

  // call the procedure defined for the type for doing the define
  if ((*v_Var->i_Methods->i_DefineProc)(v_Var, a_Cursor, a_Position,
        v_Param) < 0) {
    Py_DECREF(v_Var);
    return NULL;
  }

  return v_Var;
}


//-----------------------------------------------------------------------------
// Variable_Bind()
//   Allocate a variable and bind it to the given statement.
//-----------------------------------------------------------------------------
static int Variable_Bind(
  udt_Variable *a_Var,			// variable to bind
  udt_Cursor *a_Cursor,			// cursor to bind variable to
  PyObject *a_Name)			// name to bind to
{
  if (a_Var->i_BindHandle)
    return 0;
  return (*a_Var->i_Methods->i_BindProc)(a_Var, a_Cursor, a_Name);
}


//-----------------------------------------------------------------------------
// Variable_GetSingleValue()
//   Return the value of the variable at the given position.
//-----------------------------------------------------------------------------
static PyObject *Variable_GetSingleValue(
  udt_Variable *a_Var,			// variable to get the value for
  unsigned a_ArrayPos)			// array position
{
  // ensure we do not exceed the number of allocated elements
  if (a_ArrayPos >= a_Var->i_AllocatedElements) {
    PyErr_SetString(PyExc_IndexError,
        "Variable_GetSingleValue: array size exceeded");
    return NULL;
  }

  // check for a NULL value
  if (a_Var->i_Indicator[a_ArrayPos] == OCI_IND_NULL) {
    Py_INCREF(Py_None);
    return Py_None;
  }

  return (*a_Var->i_Methods->i_GetValueProc)(a_Var, a_ArrayPos);
}


//-----------------------------------------------------------------------------
// Variable_GetArrayValue()
//   Return the value of the variable as a PL/SQL array.
//-----------------------------------------------------------------------------
static PyObject *Variable_GetArrayValue(
  udt_Variable *a_Var)			// variable to get the value for
{
  PyObject *v_Value, *v_SingleValue;
  int i;

  v_Value = PyList_New(a_Var->i_ActualElements);
  if (!v_Value)
    return NULL;

  for (i=0; i < a_Var->i_ActualElements; i++) {
    v_SingleValue = Variable_GetSingleValue(a_Var, i);
    if (!v_SingleValue) {
      Py_DECREF(v_Value);
      return NULL;
    }
    PyList_SET_ITEM(v_Value, i, v_SingleValue);
  }

  return v_Value;
}


//-----------------------------------------------------------------------------
// Variable_GetValue()
//   Return the value of the variable.
//-----------------------------------------------------------------------------
static PyObject *Variable_GetValue(
  udt_Variable *a_Var,			// variable to get the value for
  unsigned a_ArrayPos)			// array position
{
  if (a_Var->i_IsArray)
    return Variable_GetArrayValue(a_Var);
  return Variable_GetSingleValue(a_Var, a_ArrayPos);
}


//-----------------------------------------------------------------------------
// Variable_SetSingleValue()
//   Set a single value in the variable.
//-----------------------------------------------------------------------------
static int Variable_SetSingleValue(
  udt_Variable *a_Var,			// variable to set value for
  unsigned a_ArrayPos,			// array position
  PyObject *a_Value)			// value to set
{
  // ensure we do not exceed the number of allocated elements
  if (a_ArrayPos >= a_Var->i_AllocatedElements) {
    PyErr_SetString(PyExc_IndexError,
        "Variable_SetSingleValue: array size exceeded");
    return -1;
  }

  // check for a NULL value
  if (a_Value == Py_None) {
    a_Var->i_Indicator[a_ArrayPos] = OCI_IND_NULL;
    return 0;
  }

  a_Var->i_Indicator[a_ArrayPos] = OCI_IND_NOTNULL;
  return (*a_Var->i_Methods->i_SetValueProc)(a_Var, a_ArrayPos, a_Value);
}


//-----------------------------------------------------------------------------
// Variable_SetArrayValue()
//   Set all of the array values for the variable.
//-----------------------------------------------------------------------------
static int Variable_SetArrayValue(
  udt_Variable *a_Var,			// variable to set value for
  PyObject *a_Value)			// value to set
{
  unsigned v_NumElements;
  int i;

  // ensure we have an array to set
  if (!PyList_Check(a_Value)) {
    PyErr_SetString(PyExc_TypeError, "expecting array data");
    return -1;
  }

  // ensure we haven't exceeded the number of allocated elements
  v_NumElements = PyList_GET_SIZE(a_Value);
  if (v_NumElements >= a_Var->i_AllocatedElements) {
    PyErr_SetString(PyExc_IndexError,
        "Variable_SetArrayValue: array size exceeded");
    return -1;
  }

  // set all of the values
  a_Var->i_ActualElements = v_NumElements;
  for (i = 0; i < a_Var->i_ActualElements; i++) {
    if (Variable_SetSingleValue(a_Var, i, PyList_GET_ITEM(a_Value, i)) < 0)
      return -1;
  }
  return 0;
}


//-----------------------------------------------------------------------------
// Variable_SetValue()
//   Set the value of the variable.
//-----------------------------------------------------------------------------
static int Variable_SetValue(
  udt_Variable *a_Var,			// variable to set
  unsigned a_ArrayPos,			// array position
  PyObject *a_Value)			// value to set
{
  if (a_Var->i_IsArray)
    return Variable_SetArrayValue(a_Var, a_Value);
  return Variable_SetSingleValue(a_Var, a_ArrayPos, a_Value);
}


//-----------------------------------------------------------------------------
// Variable_Copy()
//   Copy the contents of the source variable to the destination variable.
//-----------------------------------------------------------------------------
static PyObject *Variable_Copy(
  udt_Variable *a_Var,			// variable to set
  PyObject *a_Args)			// arguments
{
  unsigned v_SourcePos, v_TargetPos;
  udt_Variable *v_Var;

  // parse arguments; verify that copy is possible
  if (!PyArg_ParseTuple(a_Args, "Oii", &v_Var, &v_SourcePos, &v_TargetPos))
    return NULL;
  if (a_Var->ob_type != v_Var->ob_type) {
printf("Source type: %s\n", v_Var->ob_type->tp_name);
printf("Target type: %s\n", a_Var->ob_type->tp_name);
    PyErr_SetString(g_ProgrammingErrorException,
        "source variable must have the same type as the target variable");
    return NULL;
  }
  if (!a_Var->i_Methods->i_CopyProc) {
    PyErr_SetString(g_ProgrammingErrorException,
        "variable does not support the copy method");
    return NULL;
  }

  // ensure array positions are not violated
  if (v_SourcePos >= v_Var->i_AllocatedElements) {
    PyErr_SetString(PyExc_IndexError,
        "Variable_Copy: source array size exceeded");
    return NULL;
  }
  if (v_TargetPos >= a_Var->i_AllocatedElements) {
    PyErr_SetString(PyExc_IndexError,
        "Variable_Copy: target array size exceeded");
    return NULL;
  }

  // handle null case directly
  if (v_Var->i_Indicator[v_SourcePos] == OCI_IND_NULL)
    a_Var->i_Indicator[v_TargetPos] = OCI_IND_NULL;

  // otherwise, call copy method
  else {
    a_Var->i_Indicator[v_TargetPos] = OCI_IND_NOTNULL;
    if ((*a_Var->i_Methods->i_CopyProc)(a_Var, v_Var, v_SourcePos,
      v_TargetPos) < 0)
      return NULL;
  }

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Variable_ExternalSetValue()
//   Set the value of the variable at the given position.
//-----------------------------------------------------------------------------
static PyObject *Variable_ExternalSetValue(
  udt_Variable *a_Var,			// variable to set
  PyObject *a_Args)			// arguments
{
  PyObject *v_Value;
  unsigned v_Pos;

  if (!PyArg_ParseTuple(a_Args, "iO", &v_Pos, &v_Value))
    return NULL;
  if (Variable_SetValue(a_Var, v_Pos, v_Value) < 0)
    return NULL;

  Py_INCREF(Py_None);
  return Py_None;
}


//-----------------------------------------------------------------------------
// Variable_ExternalGetValue()
//   Return the value of the variable at the given position.
//-----------------------------------------------------------------------------
static PyObject *Variable_ExternalGetValue(
  udt_Variable *a_Var,			// variable to set
  PyObject *a_Args)			// arguments
{
  unsigned v_Pos = 0;

  if (!PyArg_ParseTuple(a_Args, "|i", &v_Pos))
    return NULL;
  return Variable_GetValue(a_Var, v_Pos);
}


//-----------------------------------------------------------------------------
// Variable_GetAttr()
//   Retrieve an attribute on the variable object.
//-----------------------------------------------------------------------------
static PyObject *Variable_GetAttr(
  udt_Variable *a_Var,			// variable object
  PyObject *a_Name)			// name of attribute
{
  return Py_FindMethod(g_VariableMethods, (PyObject*) a_Var,
      PyString_AS_STRING(a_Name));
}


