//-----------------------------------------------------------------------------
// cx_Oracle.c
//   Shared library for use by Python.
//-----------------------------------------------------------------------------

#include <Python.h>
#include <time.h>
#include <oci.h>

//-----------------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------------
static PyObject *g_WarningException = NULL;
static PyObject *g_ErrorException = NULL;
static PyObject *g_InterfaceErrorException = NULL;
static PyObject *g_DatabaseErrorException = NULL;
static PyObject *g_DataErrorException = NULL;
static PyObject *g_OperationalErrorException = NULL;
static PyObject *g_IntegrityErrorException = NULL;
static PyObject *g_InternalErrorException = NULL;
static PyObject *g_ProgrammingErrorException = NULL;
static PyObject *g_NotSupportedErrorException = NULL;
static PyObject *g_ModuleDictionary = NULL;


//-----------------------------------------------------------------------------
// SetException()
//   Create an exception and set it in the provided dictionary.
//-----------------------------------------------------------------------------
static int SetException(
  PyObject *a_Dictionary,		// dictionary to modify
  PyObject **a_Exception,		// exception to create
  char *a_Name,				// name of the exception
  PyObject *a_BaseException)		// exception to base exception on
{
  char v_Name[100];

  sprintf(v_Name, "cx_Oracle.%s", a_Name);
  *a_Exception = PyErr_NewException(v_Name, a_BaseException, NULL);
  if (!*a_Exception)
    return -1;
  return PyDict_SetItemString(a_Dictionary, a_Name, *a_Exception);
}


#include "Connection.c"


//-----------------------------------------------------------------------------
// Connect()
//   Attempt to connect to the database using the username, password and TNS
// entry.
//-----------------------------------------------------------------------------
static PyObject* Connect(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  const char *v_UserName, *v_Password, *v_TNSEntry;
  unsigned v_UserNameLen, v_PasswordLen, v_TNSEntryLen;

  // parse arguments
  v_Password = v_TNSEntry = NULL;
  v_PasswordLen = v_TNSEntryLen = 0;
  if (!PyArg_ParseTuple(a_Args, "s#|s#s#", &v_UserName, &v_UserNameLen,
      &v_Password, &v_PasswordLen, &v_TNSEntry, &v_TNSEntryLen))
    return NULL;

  // if password or TNS entry not filled in, parse the connect string
  if (!v_Password || !v_TNSEntry) {
    v_TNSEntry = strchr(v_UserName, '@');
    if (v_TNSEntry) {
      v_TNSEntry++;
      v_TNSEntryLen = strlen(v_TNSEntry);
      v_UserNameLen -= v_TNSEntryLen + 1;
    }
    v_Password = strchr(v_UserName, '/');
    if (v_Password) {
      v_Password++;
      v_PasswordLen = strlen(v_Password);
      if (v_TNSEntryLen > 0)
        v_PasswordLen -= v_TNSEntryLen + 1;
      v_UserNameLen -= v_PasswordLen + 1;
    }
  }

  // create new service object
  return Connection_Connect(v_UserName, v_UserNameLen, v_Password,
      v_PasswordLen, v_TNSEntry, v_TNSEntryLen);
}


//-----------------------------------------------------------------------------
// Attach()
//   Attempt to attach to an existing handle (as created by PowerBuilder, for
// example).
//-----------------------------------------------------------------------------
static PyObject* Attach(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  OCISvcCtx *v_Handle;

  // parse arguments
  if (!PyArg_ParseTuple(a_Args, "i", &v_Handle))
    return NULL;

  // create new service object
  return Connection_Attach(v_Handle);
}


//-----------------------------------------------------------------------------
// Date()
//   Returns a date value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* Date(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  int v_Year, v_Month, v_Day;

  // parse arguments
  if (!PyArg_ParseTuple(a_Args, "iii", &v_Year, &v_Month, &v_Day))
    return NULL;

  return ExternalDateTimeVar_New(v_Year, v_Month, v_Day, 0, 0, 0);
}


//-----------------------------------------------------------------------------
// Time()
//   Returns a time value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* Time(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  int v_Hour, v_Minute, v_Second;

  // parse arguments
  if (!PyArg_ParseTuple(a_Args, "iii", &v_Hour, &v_Minute, &v_Second))
    return NULL;

  return ExternalDateTimeVar_New(0, 0, 0, v_Hour, v_Minute, v_Second);
}


//-----------------------------------------------------------------------------
// Timestamp()
//   Returns a date value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* Timestamp(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  int v_Year, v_Month, v_Day, v_Hour, v_Minute, v_Second;

  // parse arguments
  if (!PyArg_ParseTuple(a_Args, "iiiiii", &v_Year, &v_Month, &v_Day, &v_Hour,
      &v_Minute, &v_Second))
    return NULL;

  return ExternalDateTimeVar_New(v_Year, v_Month, v_Day, v_Hour, v_Minute,
      v_Second);
}


//-----------------------------------------------------------------------------
// DateFromTicks()
//   Returns a date value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* DateFromTicks(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  struct tm *v_Time;
  time_t v_Ticks;

  // parse arguments
  if (!PyArg_ParseTuple(a_Args, "i", &v_Ticks))
    return NULL;

  v_Time = localtime(&v_Ticks);
  if (v_Time == NULL)
    return PyErr_SetFromErrno(g_DataErrorException);
  return ExternalDateTimeVar_New(v_Time->tm_year + 1900, v_Time->tm_mon + 1,
      v_Time->tm_mday, 0, 0, 0);
}


//-----------------------------------------------------------------------------
// TimeFromTicks()
//   Returns a time value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* TimeFromTicks(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  struct tm *v_Time;
  time_t v_Ticks;

  // parse arguments
  if (!PyArg_ParseTuple(a_Args, "i", &v_Ticks))
    return NULL;

  v_Time = localtime(&v_Ticks);
  if (v_Time == NULL)
    return PyErr_SetFromErrno(g_DataErrorException);
  return ExternalDateTimeVar_New(0, 0, 0, v_Time->tm_hour, v_Time->tm_min,
      v_Time->tm_sec);
}


//-----------------------------------------------------------------------------
// TimestampFromTicks()
//   Returns a date value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* TimestampFromTicks(
  PyObject* a_Self,			// passthrough argument
  PyObject* a_Args)			// arguments to function
{
  struct tm *v_Time;
  time_t v_Ticks;

  // parse arguments
  if (!PyArg_ParseTuple(a_Args, "i", &v_Ticks))
    return NULL;

  v_Time = localtime(&v_Ticks);
  if (v_Time == NULL)
    return PyErr_SetFromErrno(g_DataErrorException);
  return ExternalDateTimeVar_New(v_Time->tm_year + 1900, v_Time->tm_mon + 1,
      v_Time->tm_mday, v_Time->tm_hour, v_Time->tm_min, v_Time->tm_sec);
}


//-----------------------------------------------------------------------------
//   Declaration of methods supported by this module
//-----------------------------------------------------------------------------
static PyMethodDef g_ModuleMethods[] = {
  { "connect", Connect, METH_VARARGS },
  { "attach", Attach, METH_VARARGS },
  { "Date", Date, METH_VARARGS },
  { "Time", Time, METH_VARARGS },
  { "Timestamp", Timestamp, METH_VARARGS },
  { "DateFromTicks", DateFromTicks, METH_VARARGS },
  { "TimeFromTicks", TimeFromTicks, METH_VARARGS },
  { "TimestampFromTicks", TimestampFromTicks, METH_VARARGS },
  { NULL, NULL }
};

//-----------------------------------------------------------------------------
// initcx_Oracle()
//   Initialization routine for the shared libary.
//-----------------------------------------------------------------------------
void initcx_Oracle(void)
{
  PyObject *v_Module;

  // initialize types
  g_ConnectionType.ob_type = &PyType_Type;
  g_CursorType.ob_type = &PyType_Type;
  g_StringVarType.ob_type = &PyType_Type;
  g_RowidVarType.ob_type = &PyType_Type;
  g_BinaryVarType.ob_type = &PyType_Type;
  g_LongStringVarType.ob_type = &PyType_Type;
  g_LongBinaryVarType.ob_type = &PyType_Type;
  g_NumberVarType.ob_type = &PyType_Type;
  g_DateTimeVarType.ob_type = &PyType_Type;
  g_LobVarType.ob_type = &PyType_Type;
  g_ExternalLobVarType.ob_type = &PyType_Type;
  g_ExternalDateTimeVarType.ob_type = &PyType_Type;

  // initialize module and retrieve the dictionary
  v_Module = Py_InitModule("cx_Oracle", g_ModuleMethods);
  if (!v_Module)
    return;
  g_ModuleDictionary = PyModule_GetDict(v_Module);
  if (!g_ModuleDictionary)
    return;

  // create exception object and add it to the dictionary
  if (SetException(g_ModuleDictionary, &g_WarningException, "Warning",
      PyExc_StandardError) < 0)
    return;
  if (SetException(g_ModuleDictionary, &g_ErrorException, "Error",
      PyExc_StandardError) < 0)
    return;
  if (SetException(g_ModuleDictionary, &g_InterfaceErrorException,
      "InterfaceError", g_ErrorException) < 0)
    return;
  if (SetException(g_ModuleDictionary, &g_DatabaseErrorException,
      "DatabaseError", g_ErrorException) < 0)
    return;
  if (SetException(g_ModuleDictionary, &g_DataErrorException, "DataError",
      g_DatabaseErrorException) < 0)
    return;
  if (SetException(g_ModuleDictionary, &g_OperationalErrorException,
      "OperationalError", g_DatabaseErrorException) < 0)
    return;
  if (SetException(g_ModuleDictionary, &g_IntegrityErrorException,
      "IntegrityError", g_DatabaseErrorException) < 0)
    return;
  if (SetException(g_ModuleDictionary, &g_InternalErrorException,
      "InternalError", g_DatabaseErrorException) < 0)
    return;
  if (SetException(g_ModuleDictionary, &g_ProgrammingErrorException,
      "ProgrammingError", g_DatabaseErrorException) < 0)
    return;
  if (SetException(g_ModuleDictionary, &g_NotSupportedErrorException,
      "NotSupportedError", g_DatabaseErrorException) < 0)
    return;

  // create the basic data types for setting input sizes
  if (PyDict_SetItemString(g_ModuleDictionary, "STRING",
      (PyObject*) &g_StringVarType) < 0)
    return;
  if (PyDict_SetItemString(g_ModuleDictionary, "NUMBER",
      (PyObject*) &g_NumberVarType) < 0)
    return;
  if (PyDict_SetItemString(g_ModuleDictionary, "DATETIME",
      (PyObject*) &g_ExternalDateTimeVarType) < 0)
    return;
  if (PyDict_SetItemString(g_ModuleDictionary, "BINARY",
      (PyObject*) &g_BinaryVarType) < 0)
    return;
  if (PyDict_SetItemString(g_ModuleDictionary, "ROWID",
      (PyObject*) &g_RowidVarType) < 0)
    return;
  if (PyDict_SetItemString(g_ModuleDictionary, "LONG_STRING",
      (PyObject*) &g_LongStringVarType) < 0)
    return;
  if (PyDict_SetItemString(g_ModuleDictionary, "LONG_BINARY",
      (PyObject*) &g_LongBinaryVarType) < 0)
    return;

  // create constants required by Python DB API 2.0
  if (PyDict_SetItemString(g_ModuleDictionary, "apilevel",
      PyString_FromString("2.0")) < 0)
    return;
  if (PyDict_SetItemString(g_ModuleDictionary, "threadsafety",
      PyInt_FromLong(2)) < 0)
    return;
  if (PyDict_SetItemString(g_ModuleDictionary, "paramstyle",
      PyString_FromString("named")) < 0)
    return;

  // add version for easier support
  if (PyDict_SetItemString(g_ModuleDictionary, "version",
      PyString_FromString("2.5a")) < 0)
    return;

}

